"use strict";

this.EXPORTED_SYMBOLS = ["UnMHTAdditionalDialog"];

(function(module) {

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
let { arDialog, arDialogClass } = Cu.import("resource://unmht/modules/base/arDialog.jsm", {});
let { arPrototype } = Cu.import("resource://unmht/modules/base/arPrototype.jsm", {});
let { UnMHTSave, UnMHTSaveResult } = Cu.import("resource://unmht/modules/UnMHTSave.jsm", {});

/**
 * 追加ファイルダイアログクラス
 *
 * 結果 {Promise.<Array.<string>>}
 * @resolves 追加の URI
 * @rejects {UnMHTSaveResult}
 *            UnMHTSaveResult.CANCEL
 *
 * @constructor
 * @param   {nsIDOMWindow} window
 *          対象のウィンドウ
 * @extends    arDialogClass
 * @implements arIDialog
 */
function UnMHTAdditionalDialogClass(window) {
  arDialogClass.call(this, window);
}
UnMHTAdditionalDialogClass.prototype = Object.freeze({
  /* ==== arIDialog ==== */

  /**
   * ダイアログを OK したイベント
   *
   * @returns {boolean}
   *          ダイアログを閉じるか
   */
  onDialogAccept: function() {
    try {
      this.resolve(UnMHTSave.parseAdditionalURIs(this.$("add_uris").value));
    } catch (e) {
      Cu.reportError(e + "\n" + e.stack);
      throw e;
    }

    return true;
  },

  /**
   * ダイアログをキャンセルしたイベント
   *
   * @returns {boolean}
   *          ダイアログを閉じるか
   */
  onDialogCancel: function() {
    try {
      this.reject(UnMHTSaveResult.CANCEL);
    } catch (e) {
      Cu.reportError(e + "\n" + e.stack);
      throw e;
    }

    return true;
  }
});
arPrototype.extend(UnMHTAdditionalDialogClass, arDialogClass);

/**
 * 追加ファイルダイアログ
 * @type {arDialog}
 */
let UnMHTAdditionalDialog = new arDialog(UnMHTAdditionalDialogClass, {
  url: "chrome://unmht/content/additional.xul",
  features: "chrome,titlebar,centerscreen,resizable",
  single: false,
  usePromise: true
});

module.UnMHTAdditionalDialog = UnMHTAdditionalDialog;

})(this);
