"use strict";

this.EXPORTED_SYMBOLS = ["UnMHTFennecPrefPage"];

(function(module) {

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
let { Services } = Cu.import("resource://gre/modules/Services.jsm", {});
let { Task } = Cu.import("resource://gre/modules/Task.jsm", {});
let { XPCOMUtils } = Cu.import("resource://gre/modules/XPCOMUtils.jsm", {});
let { arAlertUtils } = Cu.import("resource://unmht/modules/base/arAlertUtils.jsm", {});
let { arEvent } = Cu.import("resource://unmht/modules/base/arEvent.jsm", {});
let { arScopeManager } = Cu.import("resource://unmht/modules/base/arScoped.jsm", {});
let { arShutdownEvent } = Cu.import("resource://unmht/modules/base/arShutdownEvent.jsm", {});
let { UnMHTPref } = Cu.import("resource://unmht/modules/UnMHTPref.jsm", {});
let { UnMHTScheme } = Cu.import("resource://unmht/modules/UnMHTScheme.jsm", {});
let { TEXT } = Cu.import("resource://unmht/modules/UnMHTStringBundle.jsm", {});

let lazy = {};
XPCOMUtils.defineLazyModuleGetter(lazy, "UnMHTCache",
                                  "resource://unmht/modules/UnMHTCache.jsm");

/**
 * アプリケーション起動以外のスタートアップ時に
 * タブにある MHT ファイルを展開する
 *
 * @class
 * @implements nsIObserver
 * @implements arIStartupEventListener
 */
let UnMHTFennecPrefPage = Object.freeze({
  QueryInterface:    XPCOMUtils.generateQI([Ci.nsIObserver]),

  /* ==== nsIObserver ==== */

  /**
   * 通知を受けたイベント
   *
   * @param   {nsISupports} subject
   *          変更やアクションのあったオブジェクト
   * @param   {string} topic
   *          イベントの種類
   * @param   {string} data
   *          イベントのデータ
   */
  observe: function(subject, topic, data) {
    if (topic == "addon-options-displayed") {
      let targetDocument = subject;
      try {
        this._onAddonOptionsDisplayed(targetDocument);
      } catch (e) {
        Cu.reportError(e + "\n" + e.stack);
      }
    }
  },

  /* ==== arIStartupEventListener ==== */

  /**
   * スタートアップ時のイベント
   *
   * @param   {Object} data
   *          Bootstrap データ
   * @param   {number} reason
   *          理由
   */
  onStartup: function(data, reason) {
    Services.obs.addObserver(this, "addon-options-displayed", false);
    arShutdownEvent.addFunc((data, reason) => {
      Services.obs.removeObserver(this, "addon-options-displayed");
    });
  },

  /* ==== private ==== */

  /**
   * 設定が表示されたイベント
   *
   * @param   {nsIDOMDocument} targetDocument
   *          対象のドキュメント
   */
  _onAddonOptionsDisplayed: function(targetDocument) {
    let targetWindow = targetDocument.defaultView;
    let $ = id => targetDocument.getElementById(id);

    arEvent.add(targetDocument, targetWindow, "beforeunload", () => {
      arScopeManager.leave(targetDocument);
    }, true);

    let overrides = [
      ["save.contents.video", "ask", "save"],
      ["save.contents.embed", "ask", "save"],
      ["save.quick.dup.action", "ask", "rename"],
    ];
    for (let [name, from, to] of overrides) {
      let value = UnMHTPref.getPref(name);
      if (value == from) {
        UnMHTPref.setPref(name, to);
      }
    }

    let clear = $("ui-open.cache.clear");
    arEvent.add(targetDocument, clear, "command", Task.async(function*() {
      let window = Services.wm.getMostRecentWindow("navigator:browser");
      let title = TEXT("name");
      let message = TEXT("cache.clear.confirm");
      let result = yield arAlertUtils.confirm(window, title, message);

      if (!result) {
        return;
      }

      lazy.UnMHTCache.removeOld(0);

      arAlertUtils.toast(window, TEXT("cache.clear.done"), "short");
    }), true);

    arEvent.add(targetDocument, $("ui-open.cache.show"), "command", () => {
      let window = Services.wm.getMostRecentWindow("navigator:browser");
      window.BrowserApp.addTab(UnMHTScheme.createCacheURI());
    }, true);

    let reset = $("ui-reset");
    arEvent.add(targetDocument, reset, "command", Task.async(function*() {
      let window = Services.wm.getMostRecentWindow("navigator:browser");
      let title = TEXT("name");
      let message = TEXT("pref.reset.confirm");
      let result = yield arAlertUtils.confirm(window, title, message);
      if (!result) {
        return;
      }

      for (let name of UnMHTPref.names) {
        UnMHTPref.clearPref(name);
      }
    }), true);
  }
});

module.UnMHTFennecPrefPage = UnMHTFennecPrefPage;

})(this);
