"use strict";

this.EXPORTED_SYMBOLS = ["UnMHTLocationBar"];

(function(module) {

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
let { Services } = Cu.import("resource://gre/modules/Services.jsm", {});
let { setTimeout } = Cu.import("resource://gre/modules/Timer.jsm", {});
let { XPCOMUtils } = Cu.import("resource://gre/modules/XPCOMUtils.jsm", {});
let { arWindowUtils } = Cu.import("resource://unmht/modules/base/arWindowUtils.jsm", {});
let { arWindowWatcher } = Cu.import("resource://unmht/modules/base/arWindowWatcher.jsm", {});
let { TEXT } = Cu.import("resource://unmht/modules/UnMHTStringBundle.jsm", {});
let { UnMHTSaveState, UnMHTSaveStateUtils } = Cu.import("resource://unmht/modules/UnMHTSaveState.jsm", {});
let { UnMHTScheme } = Cu.import("resource://unmht/modules/UnMHTScheme.jsm", {});

let lazy = {};
XPCOMUtils.defineLazyModuleGetter(lazy, "arFormatter",
                                  "resource://unmht/modules/base/arFormatter.jsm");
XPCOMUtils.defineLazyModuleGetter(lazy, "arMIMEParser",
                                  "resource://unmht/modules/base/arMIMEParser.jsm");
XPCOMUtils.defineLazyModuleGetter(lazy, "arWindowOverlay",
                                  "resource://unmht/modules/base/arWindowOverlay.jsm");
XPCOMUtils.defineLazyModuleGetter(lazy, "UnMHTCache",
                                  "resource://unmht/modules/UnMHTCache.jsm");
XPCOMUtils.defineLazyServiceGetter(lazy, "ClipboardHelper",
                                   "@mozilla.org/widget/clipboardhelper;1",
                                   "nsIClipboardHelper");

/**
 * 情報パネル
 *
 * @constructor
 */
function UnMHTInfoPanel() {
  /**
   * 情報パネル
   * @type {nsIDOMXULElement}
   */
  this.node = null;

  /**
   * 情報パネルに表示されている本来のアドレス
   * @type {string}
   */
  this.location = "";

  /**
   * 情報パネルに表示されている保存元
   * @type {string}
   */
  this.savedFrom = "";

  /**
   * 保存元がリンク可能か
   * @type {boolean}
   */
  this.linkable = false;

  Object.seal(this);
}
UnMHTInfoPanel.prototype = Object.freeze({
  /**
   * 情報パネルを表示する
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   * @param   {UnMHTExtractFileInfo} eFileInfo
   *          展開情報
   * @param   {arMIMEPart} part
   *          対象のパート
   * @param   {nsIDOMXULElement} icon
   *          アンカーにするアイコン
   */
  show: function(window, eFileInfo, part, icon) {
    let document = window.document;
    let $ = id => document.getElementById(id);

    this.location = eFileInfo.original;
    this.savedFrom = "";
    this.linkable = false;

    this.hide();
    this.node = $("unmht-info-panel");

    this.node.hidden = false;
    $("unmht-info-panel").style.width = $("urlbar").boxObject.width + "px";
    $("unmht-info-panel-original").value = this.location;

    if (eFileInfo.startPart &&
        eFileInfo.startPart.contentLocation) {
      this.savedFrom
        = part.contentLocation ||
        eFileInfo.startPart.contentLocation;

      $("unmht-info-panel-row-saved-from").hidden = false;
      $("unmht-info-panel-saved-from").value = this.savedFrom;

      this.linkable = /^(?:https?|s?ftp):/.test(this.savedFrom);

      if (this.linkable) {
        $("unmht-info-panel-saved-from").className = "text-link";
      } else {
        $("unmht-info-panel-saved-from").className = "";
      }
    } else {
      $("unmht-info-panel-row-saved-from").hidden = true;
    }

    function serializeMailBox(box, count, mbox) {
      if (mbox.isGroup) {
        if (count.value < count.max) {
          if (mbox.name) {
            let node = document.createTextNode(mbox.name + ": ");
            box.appendChild(node);
          }
        }

        serializeMailBoxList(box, count, mbox.list);

        if (count.value < count.max) {
          let node = document.createTextNode(";");
          box.appendChild(node);
        }

        return;
      }

      if (count.value < count.max) {
        let nowrapBox = document.createElementNS("http://www.w3.org/1999/xhtml", "span");
        nowrapBox.style.whiteSpace = "nowrap";
        box.appendChild(nowrapBox);

        if (mbox.name) {
          let node = document.createTextNode(mbox.name + " ");
          nowrapBox.appendChild(node);
        }

        let node = document.createTextNode("<");
        nowrapBox.appendChild(node);

        node = document.createElementNS("http://www.w3.org/1999/xhtml", "a");
        node.style.color = "-moz-nativehyperlinktext";
        node.style.cursor = "pointer";
        node.href = "mailto:" + mbox.addr;
        node.appendChild(document.createTextNode(mbox.addr));
        nowrapBox.appendChild(node);

        node = document.createTextNode(">");
        nowrapBox.appendChild(node);
      }
      count.value ++;
    }

    function serializeMailBoxList(box, count, list) {
      let first = true;
      for (let mbox of list) {
        if (first) {
          first = false;
        } else {
          if (count.value < count.max) {
            let node = document.createTextNode(", ");
            box.appendChild(node);
          }
        }
        serializeMailBox(box, count, mbox);
      }
    }

    function serializeCount(box, count) {
      if (count.value > count.max) {
        let node = document.createTextNode(TEXT("info.fields.more",
                                                [count.value - count.max]));
        box.appendChild(node);
      }
    }

    function formatMailBox(box, text) {
      while (box.firstChild) {
        box.firstChild.remove();
      }

      let ok = false;

      let context = new lazy.arMIMEParser(text);
      context._O(() => {
        let mbox = context.mailbox();
        context.__END();

        ok = true;
        let count = { value: 0, max: 4 };
        serializeMailBox(box, count, mbox);
        serializeCount(box, count);
      });

      return ok;
    }

    function formatMailBoxList(box, text) {
      while (box.firstChild) {
        box.firstChild.remove();
      }

      let ok = false;

      let context = new lazy.arMIMEParser(text);
      context._O(() => {
        let list = context.mailbox_list();
        context.__END();

        ok = true;
        let count = { value: 0, max: 4 };
        serializeMailBoxList(box, count, list);
        serializeCount(box, count);
      });

      return ok;
    }

    function formatAddressList(box, text) {
      while (box.firstChild) {
        box.firstChild.remove();
      }

      let ok = false;

      let context = new lazy.arMIMEParser(text);
      context._O(() => {
        let list = context.address_list();
        context.__END();

        ok = true;
        let count = { value: 0, max: 4 };
        serializeMailBoxList(box, count, list);
        serializeCount(box, count);
      });

      return ok;
    }

    if (eFileInfo.topPart.hasField("from")) {
      let value = eFileInfo.topPart.getField("from");
      if (formatMailBoxList($("unmht-info-panel-from"), value)) {
        $("unmht-info-panel-row-from").hidden = false;
      } else {
        $("unmht-info-panel-row-from").hidden = true;
      }
    } else {
      $("unmht-info-panel-row-from").hidden = true;
    }

    if (eFileInfo.subject) {
      let value = eFileInfo.subject;
      $("unmht-info-panel-row-subject").hidden = false;
      $("unmht-info-panel-subject").value = value;
    } else {
      $("unmht-info-panel-row-subject").hidden = true;
    }

    if (eFileInfo.topPart.hasField("sender")) {
      let value = eFileInfo.topPart.getField("sender");
      if (formatMailBox($("unmht-info-panel-sender"), value)) {
        $("unmht-info-panel-row-sender").hidden = false;
      } else {
        $("unmht-info-panel-row-sender").hidden = true;
      }
    } else {
      $("unmht-info-panel-row-sender").hidden = true;
    }

    if (eFileInfo.topPart.hasField("reply-to")) {
      let value = eFileInfo.topPart.getField("reply-to");
      if (formatAddressList($("unmht-info-panel-reply-to"), value)) {
        $("unmht-info-panel-row-reply-to").hidden = false;
      } else {
        $("unmht-info-panel-row-reply-to").hidden = true;
      }
    } else {
      $("unmht-info-panel-row-reply-to").hidden = true;
    }

    if (eFileInfo.topPart.hasField("to")) {
      let value = eFileInfo.topPart.getField("to");
      if (formatAddressList($("unmht-info-panel-to"), value)) {
        $("unmht-info-panel-row-to").hidden = false;
      } else {
        $("unmht-info-panel-row-to").hidden = true;
      }
    } else {
      $("unmht-info-panel-row-to").hidden = true;
    }

    if (eFileInfo.topPart.hasField("cc")) {
      let value = eFileInfo.topPart.getField("cc");
      if (formatAddressList($("unmht-info-panel-cc"), value)) {
        $("unmht-info-panel-row-cc").hidden = false;
      } else {
        $("unmht-info-panel-row-cc").hidden = true;
      }
    } else {
      $("unmht-info-panel-row-cc").hidden = true;
    }

    if (eFileInfo.topPart.hasField("bcc")) {
      let value = eFileInfo.topPart.getField("bcc");
      $("unmht-info-panel-row-bcc").hidden = false;
      if (formatAddressList($("unmht-info-panel-bcc"), value)) {
        $("unmht-info-panel-row-bcc").hidden = false;
      } else {
        $("unmht-info-panel-row-bcc").hidden = true;
      }
    } else {
      $("unmht-info-panel-row-bcc").hidden = true;
    }

    if (eFileInfo.date) {
      let value = lazy.arFormatter.date(eFileInfo.date);
      $("unmht-info-panel-row-saved-date").hidden = false;
      $("unmht-info-panel-saved-date").value = value;
    } else {
      $("unmht-info-panel-row-saved-date").hidden = true;
    }

    if (eFileInfo.topPart.hasField("x-mailer")) {
      let value = eFileInfo.topPart.getField("x-mailer");
      $("unmht-info-panel-row-mailer").hidden = false;
      $("unmht-info-panel-mailer").value = value;
    } else {
      $("unmht-info-panel-row-mailer").hidden = true;
    }

    if (eFileInfo.topPart.hasField("X-UnMHT-Save-State")) {
      let state = UnMHTSaveState.UNKNOWN;

      switch (eFileInfo.topPart.getField("X-UnMHT-Save-State")) {
        case "Current-State": {
          state = UnMHTSaveState.CURRENT;
          break;
        }
        case "Current-State-Without-Script": {
          state = UnMHTSaveState.CURRENT_NOJS;
          break;
        }
        case "Original-File": {
          state = UnMHTSaveState.ORIGINAL;
          break;
        }
      }

      let saveState = UnMHTSaveStateUtils.saveStateToReadableString(state);

      if (saveState) {
        $("unmht-info-panel-row-save-state").hidden = false;
        $("unmht-info-panel-save-state").value = saveState;
      } else {
        $("unmht-info-panel-row-save-state").hidden = true;
      }
    } else {
      $("unmht-info-panel-row-save-state").hidden = true;
    }

    this.node.openPopup(icon, "bottomcenter topright");
  },

  /**
   * 破棄
   */
  destruct: function() {
    this.hide();
  },

  /**
   * 情報バネルを閉じる
   */
  hide: function() {
    if (this.node) {
      try {
        this.node.hidePopup();
      } catch (e) {
      }
      this.node = null;
    }
  },

  /**
   * 本来のアドレスをコピーする
   */
  copyLocation: function() {
    lazy.ClipboardHelper.copyString(this.location);
  },

  /**
   * 保存元を開く
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   */
  openSavedFrom: function(window) {
    if (!this.linkable) {
      return;
    }

    this.node.hidePopup();

    let tabbrowser = arWindowUtils.getTabBrowser(window);
    tabbrowser.selectedTab = tabbrowser.addTab(this.savedFrom);
  }
});

/**
 * ロケーションバー
 * 情報パネルと情報アイコン管理
 *
 * @class
 * @implements arIStartupEventListener
 * @implements arIWindowEventListener
 */
let UnMHTLocationBar = Object.freeze({
  /**
   * ロケーションバーと情報アイコンの状態を状況に合わせて切り替える
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   */
  setState: function(window) {
    this._panel.hide();

    let result = this._updateInfoIcon(window);
    if (!result) {
      return;
    }

    if (result.param.scheme == "unmht-error") {
      this._setLocationBar(window, UnMHTScheme.deErrorURI(result.location));
    } else if (result.eFileInfo &&
               result.part &&
               result.param.refName === "") {
      /* 開始パートの場合のみロケーションバーに本来の URI を表示する */
      this._setLocationBar(window, result.param.original);
    }
  },

  /**
   * 情報アイコンがクリックされたイベント
   *
   * @param   {Event} event
   *          対象のイベント
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   */
  onInfoIconClick: function(event, window) {
    let document = window.document;
    let $ = id => document.getElementById(id);

    let result = this._updateInfoIcon(window);
    if (!result) {
      return;
    }

    let icon = $("unmht-urlbar-icon");
    this._panel.show(window, result.eFileInfo, result.part, icon);

    event.stopPropagation();
  },

  /**
   * 情報パネルのコピーボタンがクリックされたイベント
   *
   * @param   {Event} event
   *          対象のイベント
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   */
  onInfoCopyClick: function(event, window) {
    this._panel.copyLocation();
  },

  /**
   * 情報パネルのコピーボタンがクリックされたイベント
   *
   * @param   {Event} event
   *          対象のイベント
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   */
  onInfoSavedFromClick: function(event, window) {
    this._panel.openSavedFrom(window);
  },

  /* ==== arIStartupEventListener ==== */

  /**
   * スタートアップ時のイベント
   *
   * @param   {Object} data
   *          Bootstrap データ
   * @param   {number} reason
   *          理由
   */
  onStartup: function(data, reason) {
    arWindowWatcher.addListener(this);
  },

  /* ==== arIWindowEventListener ==== */

  /**
   * ウィンドウが開いたイベント
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   */
  onWindowOpen: function(window) {
  },

  /**
   * ウィンドウの DOM のロードイベント
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   */
  onWindowDOMLoad: function(window) {
  },

  /**
   * ウィンドウのロードイベント
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   */
  onWindowLoad: function(window) {
    if (!arWindowUtils.isBrowser(window)) {
      return;
    }

    this._createUI(window);

    this.setState(window);

    function checkDocument(webProgress) {
      if (!("DOMWindow" in webProgress &&
            webProgress.DOMWindow &&
            "document" in webProgress.DOMWindow &&
            webProgress.DOMWindow.document)) {
        return false;
      }

      let browser = arWindowUtils.getCurrentBrowser(window);
      return webProgress.DOMWindowID == browser.webProgress.DOMWindowID;
    }

    arWindowUtils.addProgressListener(window, {
      QueryInterface: XPCOMUtils.generateQI([Ci.nsISupportsWeakReference,
                                             Ci.nsIWebProgressListener]),
      onLocationChange: function(webProgress, request, location) {
        if (!checkDocument(webProgress)) {
          return;
        }
        UnMHTLocationBar.setState(window);
      },
      onProgressChange: function(webProgress, request,
                                 curSelfProgress, maxSelfProgress,
                                 curTotalProgress, maxTotalProgress) {},
      onSecurityChange: function(webProgress, request, state) {},
      onStateChange: function(webProgress, request, flags, status) {
        if (flags & Ci.nsIWebProgressListener.STATE_STOP) {
          if (!checkDocument(webProgress)) {
            return;
          }
          UnMHTLocationBar.setState(window);
        }
      },
      onStatusChange: function(webProgress, request, status, message) {}
    });
  },

  /**
   * ウィンドウのアンロードイベント
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   */
  onWindowUnload: function(window) {
    this._panel.destruct();
  },

  /* ==== private ==== */

  /**
   * 情報パネル
   * @type {UnMHTInfoPanel}
   */
  _panel: new UnMHTInfoPanel(),

  /**
   * UI を作る
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   */
  _createUI: function(window) {
    let overlay = lazy.arWindowOverlay.get(window);

    let infoPanel = overlay.mainPopupSet.addPanel({
      id: "unmht-info-panel",
      type: "arrow",
      orient: "vertical",
      ignorekeys: "true",
      hidden: "true"
    });

    let grid = infoPanel.addGrid(null, {
      columns: [{}, { flex: 1}]
    });

    let row;

    row = grid.addRow({
      align: "center",
    });
    row.addLabel({
      value: TEXT("info.original_address"),
      style: "text-align: end;"
    });
    row.addTextBox({
      id: "unmht-info-panel-original",
      value: "",
      flex: "1",
      readonly: "true"
    });
    row.addButton({
      id: "unmht-info-panel-copy",
      label: TEXT("info.copy"),
      onclick: function(event) {
        UnMHTLocationBar.onInfoCopyClick(event, window);
      }
    });

    row = grid.addRow({
      id: "unmht-info-panel-row-saved-from",
      align: "center"
    });
    row.addLabel({
      value: TEXT("info.saved_from"),
      style: "text-align: end;"
    });
    row.addElement("description", {
      id: "unmht-info-panel-saved-from",
      value: "",
      crop: "end",
      flex: "1",
      style: "margin: 2px 6px",
      onclick: function(event) {
        UnMHTLocationBar.onInfoSavedFromClick(event, window);
      }
    });

    let fields = [
      /* 標準ヘッダ */
      ["subject", null, "Subject:"],

      ["from", null, "From:"],
      ["sender", null, "Sender:"],
      ["reply-to", null, "Reply-To:"],
      ["to", null, "To:"],
      ["cc", null, "Cc:"],
      ["bcc", null, "Bcc:"],
      ["saved-date", "saved_date", null],
      ["mailer", null, "X-Mailer:"],

      /* UnMHT 用 */
      ["save-state", "save_state", null],
    ];

    for (let [id, label_id, label] of fields) {
      row = grid.addRow({
        id: "unmht-info-panel-row-" + id,
        align: "top",
      });
      row.addLabel({
        value: (label_id
                ? TEXT("info." + label_id)
                : label),
        style: "text-align: end;"
      });
      let params = {
        id: "unmht-info-panel-" + id,
        flex: "1",
        style: "margin: 2px 6px",
      };
      row.addElement("description", params);
    }

    overlay.URLBarIcons.addIcon({
      POSITION: "first",
      id: "unmht-urlbar-icon",
      "class": "urlbar-icon",
      onclick: function(event) {
        UnMHTLocationBar.onInfoIconClick(event, window);
      }
    });
  },

  /**
   * 情報アイコンを更新する
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   * @returns {object}
   *          本来の URI、URI の情報、展開情報と対応するパート
   *          見付からなかった場合は null
   */
  _updateInfoIcon: function(window) {
    let document = window.document;
    let $ = id => document.getElementById(id);

    let icon = $("unmht-urlbar-icon");
    if (!icon) {
      return null;
    }

    let browser = arWindowUtils.getCurrentBrowser(window);
    if (!browser) {
      icon.hidden = true;
      return null;
    }

    let location = browser.currentURI.spec;
    let param = UnMHTScheme.getUnMHTURIParam(location);
    if (!param.original) {
      icon.hidden = true;
      return null;
    }

    if (param.scheme == "unmht-view-source") {
      location = UnMHTScheme.deViewSourceURI(location);
      param = UnMHTScheme.getUnMHTURIParam(location);
      if (!param.original) {
        icon.hidden = true;
        return null;
      }
    }

    let [eFileInfo, part] = lazy.UnMHTCache.getPart(location);
    if (!(eFileInfo && part)) {
      icon.hidden = true;
      return null;
    }

    icon.hidden = false;

    return {
      location: location,
      param: param,
      eFileInfo: eFileInfo,
      part: part
    };
  },

  /**
   * ロケーションバーに本来の URL を表示する
   *
   * @param   {string} original
   *          本来の URL
   */
  _setLocationBar: function(window, original) {
    let document = window.document;
    let $ = id => document.getElementById(id);

    try {
      let urlbar = $("urlbar");
      let value = window.losslessDecodeURI(Services.io.newURI(original, null, null));

      urlbar.value = value;

      /* ProgressListener の登録が XULBrowserWindow より早かった場合
       * XULBrowserWindow.onLocationChange の方が後から呼ばれ
       * ロケーションバーの表示に反映されないので
       * イベントリスナを抜けた後になるように
       * 非同期にも設定する */
      setTimeout(() => {
        urlbar.value = value;
      }, 0);
    } catch (e) {
      /* 仕様変更等で失敗する事もある
       * unmht URI scheme が表示される事自体はそれほど問題ではないので
       * その場合は何もしない */
    }
  }
});

module.UnMHTLocationBar = UnMHTLocationBar;

})(this);
