"use strict";

this.EXPORTED_SYMBOLS = ["UnMHTPref",
                         "UnMHTMailArg",
                         "UnMHTPrefJSONNames",
                         "UnMHTSaveMultiIndexStyleFileName"];

(function(module) {

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
let { OS } = Cu.import("resource://gre/modules/osfile.jsm", {});
let { Services } = Cu.import("resource://gre/modules/Services.jsm", {});
let { Task } = Cu.import("resource://gre/modules/Task.jsm", {});
let { arAppInfo } = Cu.import("resource://unmht/modules/base/arAppInfo.jsm", {});
let { arFileUtils } = Cu.import("resource://unmht/modules/base/arFileUtils.jsm", {});
let { arPrefClass } = Cu.import("resource://unmht/modules/base/arPrefClass.jsm", {});
let { toUTF8, fromUTF8 } = Cu.import("resource://unmht/modules/base/arUconv.jsm", {});
let { UnMHTSystemDirectory } = Cu.import("resource://unmht/modules/UnMHTSystemDirectory.jsm", {});

/**
 * メールの引数
 * import がループしないためにメールの引数はここで宣言する
 *
 * @class
 */
const UnMHTMailArg = Object.freeze({
  thunderbird: "-remote \"xfeDoCommand(composeMessage,subject=\'%TITLE%\',attachment=\'%URL%\')\"",
  seamonkey: "-remote \"xfeDoCommand(composeMessage,subject=\'%TITLE_NOSP%\',attachment=\'%URL%\')\"",
  macmail: "\"%TITLE%\" \"%FILE%\""
});

/**
 * デフォルトのメーラ
 * @type {string}
 */
const defaultMail = arAppInfo.isSeaMonkey ? "seamonkey" : "thunderbird";

/**
 * 設定プレフィックス
 * @type {string}
 */
const prefix = "extensions.unmht.";

/**
 * 設定と既定値
 * @type {Object}
 */
const defPrefs = Object.freeze({
  /* ---- General ---- */

  "version": "",

  "open.multiple.action": "tabs", /* tabs/index/first */

  "open.ppt.skip_warning": false,

  "open.offline": false,

  /* ---- Save ---- */

  "save.dialog.wait": 1000,

  "save.contents.video": "ask", /* save/ignore/ask */
  "save.contents.embed": "ask", /* save/ignore/ask */

  "save.header.mail.original_field": false,
  "save.header.mail.show": false,

  "save.header.last_modified": false,

  "save.insert.link": false,
  "save.insert.link.pos": "bottom", /* top/bottom */
  "save.insert.link.date": true,
  "save.insert.link.version": true,
  "save.insert.link.nolocal": false,

  "save.history.register": true,
  "save.progress.display": "dialog", /* dialog/toolbar/none */
  "save.show_errors": true,
  "save.show_option_dialog": false,

  "save.local.same_dir": false,
  "save.local.same_dir.append": "",

  "save.single.state": "current", /* current/curnojs/original */
  "save.single.state.remember": true,
  "save.single.add": false,

  /* ---- Multiple ---- */

  "save.multi.state": "current", /* current/curnojs/original */
  "save.multi.state.remember": true,
  "save.multi.format": "separated", /* separated/single
                                     * /single_noindex */

  "save.multi.dir": "",

  "save.multi.dup.action": "rename", /* replace/rename/ask */

  "save.multi.autoclose": false,
  "save.multi.tabs.autoclose": false,
  "save.multi.check_none": false,

  "save.multi.index.style": false,

  /* ---- QuickSave ---- */

  "save.quick.state": "current", /* current/curnojs/original */

  "save.quick.dir": "",

  "save.quick.dup.action": "rename", /* replace/rename/ask */

  "save.quick.use_selection": false,

  /* ---- Mail ---- */

  "save.mail.state": "current", /* current/curnojs/original */
  "save.mail.state.remember": true,

  "save.mail.app.type": defaultMail, /* thunderbird/seamonkey
                                      * /macmail/other */
  "save.mail.app": "",

  "save.mail.arg.type": defaultMail, /* thunderbird/seamonkey
                                      * /macmail/other */
  "save.mail.arg.other": toUTF8(UnMHTMailArg.thunderbird),

  "save.mail.zip": false,

  /* ---- Filename ---- */

  "save.filename": "title", /* leafname/title/custom */

  "save.filename.dup.rename.digits": 1,
  "save.filename.dup.rename.init": 2, /* 0/1/2 */

  "save.filename.sequential.digits": 1,

  "save.filename.custom": "%FILENAME%",
  "save.filename.custom.am": "a.m.",
  "save.filename.custom.pm": "p.m.",
  "save.filename.custom.limit": false,
  "save.filename.custom.limit.count": 100,
  "save.filename.custom.mail_date": false,

  "save.filename.replace.list": "[]",

  /* ---- Menu ---- */

  "menu.context.save": true,
  "menu.context.save.quick": true,
  "menu.context.save.mail": true,

  /* ---- Keyboard ---- */

  "shortcut.save.single": false,
  "shortcut.save.single.keycode": "VK_S",
  "shortcut.save.single.mods.alt": true,
  "shortcut.save.single.mods.ctrl": true,
  "shortcut.save.single.mods.meta": false,
  "shortcut.save.single.mods.shift": false,

  "shortcut.save.multi": false,
  "shortcut.save.multi.keycode": "VK_A",
  "shortcut.save.multi.mods.alt": true,
  "shortcut.save.multi.mods.ctrl": true,
  "shortcut.save.multi.mods.meta": false,
  "shortcut.save.multi.mods.shift": false,

  "shortcut.save.quick": false,
  "shortcut.save.quick.keycode": "VK_Q",
  "shortcut.save.quick.mods.alt": true,
  "shortcut.save.quick.mods.ctrl": true,
  "shortcut.save.quick.mods.meta": false,
  "shortcut.save.quick.mods.shift": false,

  "shortcut.save.overwrite": false,
  "shortcut.save.overwrite.keycode": "VK_E",
  "shortcut.save.overwrite.mods.alt": true,
  "shortcut.save.overwrite.mods.ctrl": true,
  "shortcut.save.overwrite.mods.meta": false,
  "shortcut.save.overwrite.mods.shift": false,

  /* ---- Convert ---- */

  "convert.dir.suffix": "_files",
  "convert.dir.sub": false,

  /* ---- Advanced ---- */

  "menu.all.cache": false,

  "save.insert.url": false,

  "save.mail.debug": false,

  "save.filename.replace.ie_bug.disable": false,

  /* ---- Hidden ---- */

  "integrataion.maf": true,
  "integrataion.scrapbook": true,

  "protocol.mhtml": true,

  "open.extensions": "[\"mht\",\"mhtml\",\"eml\",\"mbs\"]",
  "open.mime.types": "[\"message/rfc822\",\"message/x-mht\",\"application/mht\",\"application/x-mht\",\"application/mime\",\"application/x-mime\",\"application/x-mimearchive\",\"application/vnd.sealed.mht\",\"multipart/related\",\"text/x-mht\"]",
  "open.mime.extensions": "[[\"mhtml\",\"application/x-mht\"],[\"mht\",\"application/x-mht\"],[\"mbs\",\"application/mbox\"],[\"eml\",\"message/rfc822\"]]",
});

/**
 * JSON で格納している設定名のリスト
 * ファイルにエクスポートする際に生データを出力する
 * @type {Array.<string>}
 */
const UnMHTPrefJSONNames = Object.freeze([
  "save.filename.replace.list"
]);

const UnMHTSaveMultiIndexStyleFileName = "index_style.css";

/**
 * 設定ファイルロードイベント
 *
 * @param   {Map.<string,arPrefValue>} map
 *          設定
 * @returns {Promise.<void>}
 * @resolves 完了
 */
let onPrefFileLoad = Task.async(function*(map) {
  let name = "save.multi.index.style.content";

  if (map.has(name)) {
    let content = map.get(name);

    let filePath = yield UnMHTSystemDirectory.getSystemFilePath(UnMHTSaveMultiIndexStyleFileName);
    yield arFileUtils.createUTF8(filePath, content);
  }
});

/**
 * 設定ファイルセーブイベント
 *
 * @param {Map.<string,arPrefValue>} map
 *        設定
 * @returns {Promise.<void>}
 * @resolves 完了
 */
let onPrefFileSave = Task.async(function*(map) {
  let name = "save.multi.index.style.content";

  let filePath = yield UnMHTSystemDirectory.getSystemFilePath(UnMHTSaveMultiIndexStyleFileName);
  let content = "";
  try {
    content = yield arFileUtils.readUTF8(filePath);
  } catch (e) {
  }

  map.set(name, content);
});

/**
 * 設定マイグレーション
 */
function migrate() {
  let name = "save.multi.index.style.content";
  if (Services.prefs.getPrefType(prefix + name) == Services.prefs.PREF_STRING) {
    Task.spawn(function*() {
      let filePath = yield UnMHTSystemDirectory.getSystemFilePath(UnMHTSaveMultiIndexStyleFileName);
      let exists = yield OS.File.exists(filePath);
      if (!exists) {
        let content = fromUTF8(Services.prefs.getCharPref(prefix + name));
        yield arFileUtils.createUTF8(filePath, content);
      }
    }).catch(e => {
      Cu.reportError(e + "\n" + e.stack);
    });
  }
}

/**
 * 設定操作
 * @type {arPrefClass}
 */
let UnMHTPref = new arPrefClass(prefix, defPrefs,
                                onPrefFileLoad, onPrefFileSave);
migrate();

module.UnMHTPref = UnMHTPref;
module.UnMHTMailArg = UnMHTMailArg;
module.UnMHTPrefJSONNames = UnMHTPrefJSONNames;
module.UnMHTSaveMultiIndexStyleFileName = UnMHTSaveMultiIndexStyleFileName;

})(this);
