"use strict";

this.EXPORTED_SYMBOLS = ["UnMHTSave",
                         "UnMHTSaveUtils",
                         "UnMHTSaveResult"];

(function(module) {

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
let { OS } = Cu.import("resource://gre/modules/osfile.jsm", {});
let { Promise } = Cu.import("resource://gre/modules/Promise.jsm", {});
let { Services } = Cu.import("resource://gre/modules/Services.jsm", {});
let { Task } = Cu.import("resource://gre/modules/Task.jsm", {});
let { setTimeout } = Cu.import("resource://gre/modules/Timer.jsm", {});
let { XPCOMUtils } = Cu.import("resource://gre/modules/XPCOMUtils.jsm", {});
let { arAppInfo } = Cu.import("resource://unmht/modules/base/arAppInfo.jsm", {});
let { fromUint8Array } = Cu.import("resource://unmht/modules/base/arArrayBufferUtils.jsm", {});
let { arDOMUtils } = Cu.import("resource://unmht/modules/base/arDOMUtils.jsm", {});
let { arFileUtils } = Cu.import("resource://unmht/modules/base/arFileUtils.jsm", {});
let { arFormatter } = Cu.import("resource://unmht/modules/base/arFormatter.jsm", {});
let { arMIMEDecoder } = Cu.import("resource://unmht/modules/base/arMIMEDecoder.jsm", {});
let { arMIMEEncoder } = Cu.import("resource://unmht/modules/base/arMIMEEncoder.jsm", {});
let { arMIMEParams } = Cu.import("resource://unmht/modules/base/arMIMEParams.jsm", {});
let { arMIMEPart } = Cu.import("resource://unmht/modules/base/arMIMEPart.jsm", {});
let { arMIMEUtils } = Cu.import("resource://unmht/modules/base/arMIMEUtils.jsm", {});
let { arNetFileLoader, arNetFileLoaderStatus, arNetFileLoaderStatusMask } = Cu.import("resource://unmht/modules/base/arNetFileLoader.jsm", {});
let { arPathUtils } = Cu.import("resource://unmht/modules/base/arPathUtils.jsm", {});
let { arPseudoID } = Cu.import("resource://unmht/modules/base/arPseudoID.jsm", {});
let { arTaskUtils } = Cu.import("resource://unmht/modules/base/arTaskUtils.jsm", {});
let { arTreeWalker } = Cu.import("resource://unmht/modules/base/arTreeWalker.jsm", {});
let { arUconv } = Cu.import("resource://unmht/modules/base/arUconv.jsm", {});
let { UnMHTDOMConverter } = Cu.import("resource://unmht/modules/UnMHTDOMConverter.jsm", {});
let { UnMHTDocumentCategory } = Cu.import("resource://unmht/modules/UnMHTDocumentCategory.jsm", {});
let { UnMHTFilename } = Cu.import("resource://unmht/modules/UnMHTFilename.jsm", {});
let { UnMHTMailURLUtils } = Cu.import("resource://unmht/modules/UnMHTMailURLUtils.jsm", {});
let { UnMHTPref } = Cu.import("resource://unmht/modules/UnMHTPref.jsm", {});
let { UnMHTSaveState, UnMHTSaveStateMask } = Cu.import("resource://unmht/modules/UnMHTSaveState.jsm", {});
let { UnMHTScheme } = Cu.import("resource://unmht/modules/UnMHTScheme.jsm", {});
let { UnMHTStringBundle, TEXT } = Cu.import("resource://unmht/modules/UnMHTStringBundle.jsm", {});

let lazy = {};
XPCOMUtils.defineLazyModuleGetter(lazy, "UnMHTConfirmDialog",
                                  "resource://unmht/modules/UnMHTConfirmDialog.jsm");
XPCOMUtils.defineLazyModuleGetter(lazy, "UnMHTCustomFilename",
                                  "resource://unmht/modules/UnMHTCustomFilename.jsm");
XPCOMUtils.defineLazyModuleGetter(lazy, "UnMHTCache",
                                  "resource://unmht/modules/UnMHTCache.jsm");
XPCOMUtils.defineLazyModuleGetter(lazy, "UnMHTContentModifier",
                                  "resource://unmht/modules/UnMHTContentModifier.jsm");
XPCOMUtils.defineLazyServiceGetter(lazy, "MsgMailSession",
                                   "@mozilla.org/messenger/services/session;1",
                                   "nsIMsgMailSession");
XPCOMUtils.defineLazyServiceGetter(lazy, "DOMUtils",
                                   "@mozilla.org/inspector/dom-utils;1",
                                   "inIDOMUtils");
XPCOMUtils.defineLazyServiceGetter(lazy, "FaviconService",
                                   "@mozilla.org/browser/favicon-service;1",
                                   "nsIFaviconService");
XPCOMUtils.defineLazyServiceGetter(lazy, "ContentPolicy",
                                   "@mozilla.org/layout/content-policy;1",
                                   "nsIContentPolicy");

/**
 * ファイルの取得状況
 * @enum {number}
 */
const UnMHTSaveStatus = Object.freeze({
  /** 元のファイル (通常のファイル) - 未取得 */
  FILE_NA:               0x0001,
  /** 元のファイル (通常のファイル) - 取得中 */
  FILE_GET:              0x0011,

  /** 現在の状態 (ドキュメント) - 未取得 */
  DOC_CUR_NA:            0x0101,
  /** 現在の状態 (ドキュメント) - favicon 取得開始 */
  DOC_CUR_FAVICON_START: 0x0111,
  /** 現在の状態 (ドキュメント) - favicon 取得中 */
  DOC_CUR_FAVICON_WAIT:  0x0112,
  /** 現在の状態 (ドキュメント) - シリアライズ開始 */
  DOC_CUR_SERIAL_START:  0x0121,
  /** 現在の状態 (ドキュメント) - シリアライズ中 */
  DOC_CUR_SERIAL_WAIT:   0x0122,
  /** 現在の状態 (ドキュメント) - ファイル回収開始 */
  DOC_CUR_GATHER_START:  0x0131,
  /** 現在の状態 (ドキュメント) - ファイル回収中 */
  DOC_CUR_GATHER_WAIT:   0x0132,

  /** 元のファイル (ドキュメント) - 未取得 */
  DOC_ORI_NA:            0x0201,
  /** 元のファイル (ドキュメント) - favicon 取得開始 */
  DOC_ORI_FAVICON_START: 0x0211,
  /** 元のファイル (ドキュメント) - favicon 取得中 */
  DOC_ORI_FAVICON_WAIT:  0x0212,
  /** 元のファイル (ドキュメント) - ファイル回収開始 */
  DOC_ORI_GATHER_START:  0x0221,
  /** 元のファイル (ドキュメント) - ファイル回収中 (次は FILE_NA) */
  DOC_ORI_GATHER_WAIT:   0x0222,

  /** メールデータ取得中 (現在の状態) - メール取得開始 */
  MAIL_CUR_MAIL_START:   0x0301,
  /** メールデータ取得中 (現在の状態) - メール取得中 (次は DOC_CUR_NA) */
  MAIL_CUR_MAIL_WAIT:    0x0302,

  /*  */
  /** メールデータ取得中 (元のファイル) - メール取得開始 */
  MAIL_ORI_MAIL_START:   0x0401,
  /** メールデータ取得中 (元のファイル) - メール取得中 */
  MAIL_ORI_MAIL_WAIT:    0x0402,
  /** メールデータ取得中 (元のファイル) - ファイル未回収 */
  MAIL_ORI_NA:           0x0411,
  /** メールデータ取得中 (元のファイル) - ファイル回収開始 */
  MAIL_ORI_GATHER_START: 0x0421,
  /** メールデータ取得中 (元のファイル) - ファイル回収中 (次は IGNORE) */
  MAIL_ORI_GATHER_WAIT:  0x0422,

  /** リモートファイル取得中 - ファイル未回収 */
  REMOTE_NA:             0x0501,
  /** リモートファイル取得中 - ファイル回収開始 */
  REMOTE_GATHER_START:   0x0511,
  /** リモートファイル取得中 - ファイル回収中 (次は IGNORE) */
  REMOTE_GATHER_WAIT:    0x0512,

  /** 取得完了 */
  OK:                    0x1001,
  /** 取得失敗 */
  NG:                    0x1002,
  /** このパートを無視する */
  IGNORE:                0x1003
});

/**
 * 保存エラー
 * @enum {number}
 */
const UnMHTSaveError = Object.freeze({
  /** ファイル書き込み失敗 */
  WRITE_FILE:          0x0001,
  /** 必須ファイル取得失敗 */
  REQUIRED_FAILED:     0x0002
});

/**
 * 保存結果
 * @enum {number}
 */
const UnMHTSaveResult = Object.freeze({
  /** 成功 */
  OK:                    0x00000000,
  /** 一般エラー */
  FAIL:                  0x10000001,
  /** ユーザによるキャンセル */
  CANCEL:                0x10000002,
  /** ドキュメント追加失敗 */
  ERROR_ADD_DOCUMENT:    0x10000101,
  /** ファイル書き込み失敗 */
  ERROR_WRITE_FILE:      0x10000102,
  /** 必須ファイル取得失敗 */
  ERROR_REQUIRED_FAILED: 0x10000103
});

/**
 * 保存用ユーティリティ
 *
 * @class
 */
let UnMHTSaveUtils = Object.freeze({
  /**
   * 要素に対して定義されているスタイル名の一覧を返す
   *
   * @param   {nsIDOMElement} node
   *          対象の要素
   * @param   {string} pseudo
   *          疑似要素名
   * @param   {Map} names
   *          探す要素名 (names.has() で確認できるもの)
   * @returns {?Set.<string>}
   *          定義されているスタイル名
   */
  getDefinedStyles: function(node, pseudo, names) {
    let rules = lazy.DOMUtils.getCSSStyleRules(node, pseudo);
    if (!rules) {
      return null;
    }

    let definedStyles = new Set();

    let check = style => {
      if (!style) {
        /* MathML では node.style が undefined になる */
        return;
      }

      for (let name of Array.from(style)) {
        if (names.has(name)) {
          definedStyles.add(name);
        }
      }
    };

    check(node.style);
    for (let i = rules.Count() - 1; i >= 0; i -= 1) {
      let rule = rules.GetElementAt(i);
      if (!rule) {
        continue;
      }
      check(rule.style);
    }

    return definedStyles;
  },

  /**
   * トップパートを作成する
   *
   * @param   {string} subject
   *          ファイルの subject
   * @param   {Date} saveDate
   *          保存日時
   * @returns {arMIMEPart}
   *          トップパート
   */
  createTopPart: function(subject, saveDate) {
    let part = new arMIMEPart();
    part.isMultipart = true;
    part.boundary = arMIMEUtils.createUniqueBoundary();
    part.addField("From",
                  "<Saved by UnMHT Ver." + UnMHTPref.version + ">");
    part.addField("Subject", subject);
    part.addField("Date",
                  arMIMEEncoder.encodeDate(saveDate));
    part.addField("MIME-Version", "1.0");
    part.parts = [];

    return part;
  },

  /**
   * UnMHT が使用しないフィールドかどうか
   *
   * @param   {string} name
   *          フィールド名
   * @returns {boolean}
   *          UnMHT が使用しないフィールドかどうか
   */
  isFreeField: function(name) {
    name = name.toLowerCase();

    return !(name == "from" ||
             name == "subject" ||
             name == "date" ||
             name == "mime-version" ||
             name == "x-unmht-tabs");
  },

  /**
   * URI がロード可能かどうか調べる
   *
   * @param   {number} type
   *          nsIContentPolicy の対象の種類
   * @param   {string} location
   *          対象 URL
   * @param   {?nsIDOMDocument} targetDocument
   *          読み込み元のドキュメント
   *          null の場合は type は Ci.nsIContentPolicy.TYPE_DOCUMENT 固定
   * @returns {boolean}
   *          ロード可能か
   */
  checkPolicy: function(type, location, targetDocument) {
    if (!targetDocument) {
      type = Ci.nsIContentPolicy.TYPE_DOCUMENT;
    }

    let uri;
    try {
      uri = Services.io.newURI(location, null, null);
    } catch (e) {
      return false;
    }
    let originURI = null;
    if (targetDocument) {
      try {
        originURI = Services.io.newURI(targetDocument.documentURI, null, null);
      } catch (e) {
      }
    }

    try {
      let result = lazy.ContentPolicy.shouldLoad(type, uri, originURI,
                                                 targetDocument,
                                                 null, null);
      return (result == Ci.nsIContentPolicy.ACCEPT);
    } catch (e) {
    }

    return true;
  },

  /**
   * HTML のコメントに適した形に修正する
   *
   * @param   {string} text
   *          対象の文字列
   * @returns {string
   *          修正した文字列
   */
  toCommentBody: function(text) {
    /* コメント中に連続したハイフンは含められないので
     * 間にスペースを挿入する
     * 1 回の置換では 3 個以上の連続が消えないので 2 回行う */
    return text
      .replace(/--/g, "- -")
      .replace(/--/g, "- -");
  },

  /**
   * URL をコメントとして挿入する
   *
   * @param   {string} content
   *          ファイルの内容
   * @param   {string} location
   *          ドキュメントの URL
   * @returns {string}
   *          ファイルの内容
   * @param   {string} markerID
   *          マーカーの ID
   */
  insertComment: function(content, location, markerID) {
    if (UnMHTScheme.isUnMHTURI(location)) {
      location = UnMHTScheme.deUnMHTURI(location);
    }

    let comment
      = "<!-- Saved from "
      + this.toCommentBody(arUconv.fromUnicodeWithEntity(location, "ascii"))
      + " -->";

    if (markerID) {
      let result;
      result = UnMHTDOMConverter.appendToMarker(content, markerID,
                                                "html", "first",
                                                comment);
      if (result.ok) {
        return result.text;
      }

      result = UnMHTDOMConverter.appendToMarker(content, markerID,
                                                "head", "first",
                                                comment);
      if (result.ok) {
        return  result.text;
      }

      result = UnMHTDOMConverter.appendToMarker(content, markerID,
                                                "body", "first",
                                                comment);
      if (result.ok) {
        return result.text;
      }

      return comment + content;
    } else if (/<html[^>]*>/i.test(content)) {
      return content
        .replace(/<html[^>]*>/i,
                 matched => matched + comment);
    } else if (/<head[^>]*>/i.test(content)) {
      return content
        .replace(/<head[^>]*>/i,
                 matched => matched + comment);
    } else if (/<body[^>]*>/i.test(content)) {
      return content
        .replace(/<body[^>]*>/i,
                 matched => matched + comment);
    }

    return comment + content;
  },

  /**
   * URL をリンクとして挿入する
   *
   * @param   {string} content
   *          ファイルの内容
   * @param   {string} location
   *          ドキュメントの URL
   * @param   {Date} saveDate
   *          保存日時
   * @param   {string} markerID
   *          マーカーの ID
   * @param   {boolean} isXHTML
   *          XHTML か
   * @returns {string}
   *          ファイルの内容
   */
  insertLink: function(content, location, saveDate, markerID, isXHTML) {
    if (UnMHTScheme.isUnMHTURI(location)) {
      location = UnMHTScheme.deUnMHTURI(location);
    }

    let pos = UnMHTPref.getPref("save.insert.link.pos");

    let divStyles = [];
    if (pos == "bottom") {
      divStyles.push("clear: both");
      divStyles.push("margin-top: 2em");
    }
    divStyles.push("font-size: 0.8em");
    divStyles.push("border: 1px solid black");
    divStyles.push("background-color: white");
    divStyles.push("color: black");
    divStyles.push("padding: 0.5em");
    divStyles.push("text-align: center");

    let aStyles = [];
    aStyles.push("color: blue");
    aStyles.push("background-color: inherit");
    aStyles.push("text-decoration: underline");

    let encodedLocation
      = arUconv.fromUnicodeWithEntity(arDOMUtils.escapeEntity(location),
                                      "ascii");

    let contents = [
      "Saved from <a class=\"unmht_link_to_original\" style=\""
        + aStyles.join(";") + "\" href=\"" + encodedLocation + "\">"
        + encodedLocation + "</a>"
    ];

    if (UnMHTPref.getPref("save.insert.link.date")) {
      let encodedDate
        = arUconv.fromUnicodeWithEntity(arDOMUtils.escapeEntity(saveDate.toString()),
                                        "ascii");

      contents.push(encodedDate);
    }

    if (UnMHTPref.getPref("save.insert.link.version")) {
      contents.push("Saved by UnMHT Ver." + UnMHTPref.version);
    }

    let link
      = "<div style=\"" + divStyles.join(";") + ";\">"
      + contents.join(isXHTML ? "<br />" : "<br>")
      + "</div>";

    if (pos == "bottom") {
      content = this.modifyBody(content,
                                "",
                                link,
                                markerID);
    } else {
      content = this.modifyBody(content,
                                link + "<div style=\"position: relative\">",
                                "</div>",
                                markerID);
    }

    return content;
  },

  /**
   * ファイルを修正する
   *
   * @param   {UnMHTSaveParam} param
   *          保存情報
   * @param   {string} location
   *          対象のファイルの URI
   * @param   {Date} saveDate
   *          保存日時
   * @param   {string} markerID
   *          マーカーの ID
   */
  modifyContent: function(sParam, location, saveDate, markerID) {
    if (sParam.isHTML && !sParam.isIFrame) {
      if (UnMHTPref.getPref("save.insert.url")) {
        sParam.content = this.insertComment(sParam.content, location, markerID);
      }
      if (UnMHTPref.getPref("save.insert.link")) {
        if (!sParam.isLocal ||
            !UnMHTPref.getPref("save.insert.link.nolocal")) {
          sParam.content
            = this.insertLink(sParam.content, location, saveDate, markerID,
                              sParam.isXHTML);
        }
      }
    }
  },

  /**
   * body の最初と最後にテキストを挿入する
   *
   * @param   {string} content
   *          対象のソース
   * @param   {string} firstText
   *          body の最初に挿入するテキスト
   * @param   {string} lastText
   *          body の最後に挿入するテキスト
   * @param   {string} markerID
   *          マーカーの ID
   */
  modifyBody: function(content, firstText, lastText, markerID) {
    if (markerID) {
      let result;
      result = UnMHTDOMConverter.appendToMarker(content, markerID, "body", "first", firstText);
      if (result.ok) {
        content = result.text;
        firstText = "";
      }
      result = UnMHTDOMConverter.appendToMarker(content, markerID, "body", "last", lastText);
      if (result.ok) {
        content = result.text;
        lastText = "";
      }
    }

    if (/<body[^>]*/i.test(content)) {
      if (firstText) {
        content
          = content
          .replace(/<body[^>]*>/i,
                   matched => matched + firstText);
      }
      if (lastText) {
        content
          = content
          .replace(/<\/body>/i,
                   matched => lastText + matched);
      }
    }

    return content;
  },

  /**
   * 内容を保存するかどうかの警告を表示する
   *
   * @param   {nsIDOMWindow} window
   *          実行中のウィンドウ
   * @param   {string} type
   *          内容の種類
   * @returns {Promise.<Object>}
   * @resolves 今回の動作
   */
  showSaveContentsConfirm: Task.async(function*(window, type) {
    let arg = {
      title:    TEXT("save.contents." + type + ".title"),
      message:  (TEXT("save.contents." + type + ".message") + "\n"
                 + TEXT("save.contents." + type + ".message2")),
      accept:   TEXT("save.contents.save"),
      cancel:   TEXT("save.contents.ignore"),
      toall:    TEXT("save.contents.do_not_ask_again"),
      focus:    "yes"
    };
    let result = yield lazy.UnMHTConfirmDialog.open(window, arg);

    let action;

    if (result.action == "yes") {
      action = "save";
    } else if (result.action == "no") {
      action = "ignore";
    }

    if (result.applyToAll) {
      UnMHTPref.setPref("save.contents." + type, action);
    }

    return action;
  }),

  /**
   * MHT ファイルの Content-Location を取得する
   *
   * @param   {string} loc
   *          アドレス
   * @returns {string}
   *          アドレスが MHT ファイル中のパートを指していれば
   *          Content-Location の値
   *          それ以外ならそのまま
   */
  getMHTLocation: function(loc) {
    if (UnMHTScheme.isUnMHTURI(loc)) {
      /* MHT ファイルを再度保存する際には相対パスの基準を
       * 本来の Content-Location にする */
      let [eFileInfo, part] = lazy.UnMHTCache.getPart(loc);
      if (eFileInfo && part && part.contentLocation) {
        return part.contentLocation;
      }
    }

    return loc;
  },

  /*
   * DOCTYPE をシリアライズする
   *
   * @param   {nsIDOMDocument} targetDocument
   *          対象のドキュメント
   * @returns {string}
   *          DOCTYPE
   */
  serializeDocType: function(targetDocument) {
    return "<!DOCTYPE "
      + targetDocument.doctype.name
      + (targetDocument.doctype.publicId
         ? (" PUBLIC \""
            + arDOMUtils.escapeEntity(targetDocument.doctype.publicId)
            + "\"")
         : "")
      + (targetDocument.doctype.systemId
         ? (" \""
            + arDOMUtils.escapeEntity(targetDocument.doctype.systemId)
            + "\"")
         : "")
      + ">\n";
  },

  /*
   * DOCTYPE が XML かどうかを返す
   *
   * @param   {nsIDOMDocument} targetDocument
   *          対象のドキュメント
   * @returns {boolean}
   *          DOCTYPE が XML か
   */
  isXHTML: function(targetDocument) {
    if (targetDocument.doctype &&
        targetDocument.doctype.publicId &&
        targetDocument.doctype.publicId.startsWith("-//W3C//DTD XHTML")) {
      return true;
    }

    return false;
  },

  /**
   * メールヘッダをシリアライズする
   *
   * @param   {Array.<string,string>} mailHeaderDisplay
   *          メールに表示するヘッダの名前と値のリスト
   * @param   {boolean} isXHTML
   *          XHTML か
   * @returns {string}
   *          メールヘッダの HTML 表記
   */
  makeMailHeaderDisplay: function(mailHeaderDisplay, isXHTML) {
    let divStyles = [];
    divStyles.push("font-size: 0.8em");
    divStyles.push("background-color: white");
    divStyles.push("color: black");
    divStyles.push("padding-bottom: 1em");
    divStyles.push("text-align: left");

    return "<div style=\"" + divStyles.join(";") + ";\">"
      + mailHeaderDisplay
      .map(([name, value]) =>
           "<b>" + arDOMUtils.escapeEntity(name) + ": </b>"
           + arDOMUtils.escapeEntity(value))
      .join(isXHTML ? "<br />" : "<br>")
      + "</div>";
  },

  /**
   * メール添付ファイルをシリアライズする
   *
   * @param   {Array.<string,string>} attachments
   *          添付ファイルのファイル名と URL のリスト
   * @param   {string} location
   *          メールの URL
   * @returns {string}
   *          メール添付ファイルの HTML 表記
   */
  makeAttachmentsBox: function(attachments, location) {
    return "<div>"
      + "<hr>"
      + "<ul>"
      + attachments
      .map(([name, url]) =>
           "<li>"
           + "<a href=\"" + arDOMUtils.escapeEntity(UnMHTMailURLUtils.hideMailURL(url, location)) + "\">"
           + arDOMUtils.escapeEntity(name)
           + "</a>")
      .join("")
      + "</ul>"
      + "</div>";
  },

  /**
   * favicon の LINK 要素を持っているか
   *
   * @param   {nsIDOMElement} head
   *          HEAD 要素
   * @returns {boolean}
   *          favicon の LINK 要素を持っているか
   */
  hasFaviconNode: function(head) {
    return Array.from(head.getElementsByTagName("link"))
      .some(node => (node.hasAttribute("rel") &&
                     /icon/i.test(node.getAttribute("rel"))));
  },

  /**
   * favicon の LINK 要素を作成する
   *
   * @param   {nsIDOMDocument} targetDocument
   *          対象のドキュメント
   * @param   {string} faviconData
   *          favicon の data scheme
   * @param   {string} faviconType
   *          favicon の MIME-Type (オプショナル)
   * @returns {nsIDOMElement}
   *          LINK 要素
   */
  makeFaviconNode: function(targetDocument, faviconData, faviconType) {
    let node = targetDocument.createElement("link");
    node.setAttribute("rel", "shortcut icon");
    node.setAttribute("href", faviconData);
    if (faviconType) {
      node.setAttribute("type", faviconType);
    }
    return node;
  },

  /**
   * charset の META 要素を作成する
   *
   * @param   {nsIDOMDocument} targetDocument
   *          対象のドキュメント
   * @param   {nsIDOMElement} head
   *          HEAD 要素
   * @returns {nsIDOMElement}
   *          META 要素
   */
  makeCharsetNode: function(targetDocument, head) {
    let nodes = head.getElementsByTagName("meta");

    let oldMetas
      = Array.from(nodes)
      .filter(node => (node.hasAttribute("http-equiv") &&
                       /Content-Type/i.test(node.getAttribute("http-equiv")) &&
                       node.hasAttribute("content") &&
                       /charset=/i.test(node.getAttribute("content"))));

    let newMetas
      = Array.from(nodes)
      .filter(node => node.hasAttribute("charset"));

    let charset = targetDocument.characterSet;

    let node = targetDocument.createElement("meta");
    if (newMetas.length) {
      node.setAttribute("charset", charset);
    } else {
      node.setAttribute("http-equiv", "Content-Type");
      node.setAttribute("content", "text/html; charset=" + charset);
    }

    return [node, oldMetas.concat(newMetas)];
  },

  /**
   * BASE 要素を作成する
   *
   * @param   {nsIDOMDocument} targetDocument
   *          対象のドキュメント
   * @param   {string} baseDir
   *          href 属性
   * @returns {nsIDOMElement}
   *          BASE 要素
   */
  makeBaseNode: function(targetDocument, baseDir) {
    let node = targetDocument.createElement("base");
    node.setAttribute("href", baseDir);
    return node;
  }
});

/**
 * 保存失敗情報
 *
 * @constructor
 * @param   {UnMHTSaveError} reason
 *          エラーの理由
 * @param   {Exception} e
 *          例外
 * @param   {Array.<Tuple.<string,string>>} errorList
 *          エラーになった URL と理由のリスト
 */
function UnMHTSaveErrorInfo(reason, e, errorList) {
  /**
   * エラーの理由
   * @type {UnMHTSaveError}
   */
  this.reason = reason;

  /**
   * エラーの理由の文字列表現
   * @type {string}
   */
  this.message = "";

  /**
   * 例外
   * @type {Exception}
   */
  this.e = e;

  /**
   * エラーになった URL と理由のリスト
   * @type {Array.<Tuple.<string,string>>}
   */
  this.errorList = errorList;

  switch (reason) {
    case UnMHTSaveError.WRITE_FILE: {
      this.message = TEXT("save.error.open");
      break;
    }
    case UnMHTSaveError.REQUIRED_FAILED: {
      this.message = TEXT("save.error.required");
      break;
    }
  }

  Object.seal(this);
}
UnMHTSaveErrorInfo.prototype = Object.freeze({
});

/**
 * 保存するファイルの情報
 *
 * @constructor
 * @param   {string} loc
 *          保存するファイルの URI
 * @param   {UnMHTSaveContext} context
 *          保存コンテキスト
 * @param   {boolean} isRequired
 *          このファイルが必須か
 *          必須のファイルを取得失敗したら保存も失敗扱いになる
 * @implements arINetFileLoaderListener
 */
function UnMHTSaveParam(loc, context, isRequired) {
  /**
   * 保存元のファイルの展開情報
   * MHT ファイルから保存する場合のみ使用
   * @type {?UnMHTExtractFileInfo}
   */
  this.eFileInfo = null;

  /**
   * 保存コンテキスト
   * @type {UnMHTSaveContext}
   */
  this.saveContext = context;

  /* ---- ファイル取得 ---- */

  /**
   * 開始時間
   * @type {number}
   */
  this.startTime = 0;

  /**
   * 取得状況
   * @type {UnMHTSaveStatus}
   */
  this.status = UnMHTSaveStatus.FILE_NA;

  /**
   * 取得失敗理由
   * @type {string}
   */
  this.reason = "";

  /**
   * 取得前に分かっているファイルサイズ
   * 不明の場合 0
   * @type {number}
   */
  this.fileSize = 0;

  /**
   * 取得したファイルサイズ
   * @type {number}
   */
  this.recvSize = 0;

  /**
   * 最後に受信した時刻
   * @type {number}
   */
  this.lastRecvTime = Date.now();

  /**
   * ファイルロードコンテキスト
   * @type {arNetFileLoadContext}
   */
  this.loadContext = null;

  /**
   * ファイルローダの取得状況
   * @type {arNetFileLoaderStatus}
   */
  this.loaderStatus = 0;

  /**
   * 保存対象の URI
   * 保存の最中にリダイレクト等で変化する
   * @type {string}
   */
  this.currentLocation = loc;

  /**
   * リファラ
   * @type {string}
   */
  this.referrer = "";

  /* ---- ドキュメント取得 ---- */

  /**
   * 対象のブラウザ
   * @type {nsIXULElement}
   */
  this.browser = null;

  /**
   * 現在のドキュメント
   * @type {nsIDOMDocument}
   */
  this.targetDocument = null;

  /**
   * 選択範囲を使用するか
   * @type {boolean}
   */
  this.useSelection = false;

  /**
   * 保存状態
   * @type {UnMHTSaveState}
   */
  this.saveState = UnMHTSaveState.UNKNOWN;

  /**
   * BASE 要素を 持っているか
   * @type {boolean}
   */
  this.hasBaseElement = false;

  /**
   * ツリーウォーカ
   * @type {arTreeWalker}
   */
  this.treeWalker = null;

  /**
   * フレームの Content-ID
   * @type {Map.<nsIDOMElement,string>}
   */
  this.frameIDs = null;

  /* ---- MIME パート情報 ---- */

  /**
   * Content-Location があっても Content-ID を格納するか
   * @type {boolean}
   */
  this.storeContentID = false;

  /**
   * Content-Location に格納する値
   * 保存開始から変化しない
   * ただし MHT ファイルの場合は savedFrom と異なる
   * @type {string}
   */
  this.location = loc;

  /**
   * Content-ID に格納する値
   * @type {string}
   */
  this.cid = "";

  /**
   * MIME-Type
   * データ取得途中に変化する可能性がある
   * @type {string}
   */
  this.mimetype = "";

  /**
   * Last-Modified
   * データ取得途中に変化する可能性がある
   * @type {string}
   */
  this.lastModified = "";

  /**
   * キャラクタセット
   * データ取得途中に変化する可能性がある
   * @type {string}
   */
  this.charset = "";

  /**
   * ファイルの中身
   * @type {string}
   */
  this.content = "";

  /* ---- ファイル情報 ---- */

  /**
   * このファイルが必須か
   * 必須のファイルが失敗した場合全体の失敗扱いとなる
   * @type {boolean}
   */
  this.isRequired = isRequired;

  /**
   * ファイルをロードしたドキュメント
   * @type {nsIDOMDocument}
   */
  this.ownerDocument = null;

  /**
   * ローカルファイルか
   * @type {boolean}
   */
  this.isLocal = false;

  /**
   * ベースディレクトリ
   * MHT 中のファイルの場合 Content-Location 基準
   * @type {string}
   */
  this.baseDir = "";

  /**
   * HTML かどうか
   * コメント, リンクを入れるか
   * @type {boolean}
   */
  this.isHTML = false;

  /**
   * XHTML かどうか
   * @type {boolean}
   */
  this.isXHTML = false;

  /**
   * インラインフレームか
   * コメント, リンクを入れるか
   * @type {boolean}
   */
  this.isIFrame = false;

  /**
   * テキストかどうか
   * quoted-printable で保存するか
   * @type {boolean}
   */
  this.isText = false;

  /**
   * メールかどうか
   * @type {boolean}
   */
  this.isMail = false;

  /**
   * 保存元のアドレス
   * 保存開始から変化しない
   * @type {string}
   */
  this.savedFrom = loc;

  /* ---- メール情報 ---- */

  /**
   * 添付ファイルのファイル名と URL のリスト
   * @type {Array.<Tuple.<string,string>>}
   */
  this.attachments = null;

  /**
   * メールの URI
   * @type {string}
   */
  this.messageURI = "";

  /**
   * メールに表示するヘッダの名前と値のリスト
   * @type {Array.<Tuple.<string,string>>}
   */
  this.mailHeaderDisplay = null;

  /* ---- favicon 情報 ---- */

  /**
   * favicon の data: スキーム表現
   * @type {string}
   */
  this.faviconData = null;

  /**
   * favicon の MIME-Type
   * @type {string}
   */
  this.faviconType = null;

  let location = this.currentLocation;
  if (UnMHTScheme.isUnMHTURI(location)) {
    location = UnMHTScheme.deUnMHTURI(location);
  }
  if (location.startsWith("file://")) {
    this.isLocal = true;
  }

  if (UnMHTScheme.isUnMHTURI(this.currentLocation)) {
    let [eFileInfo, part] = lazy.UnMHTCache.getPart(this.currentLocation);
    if (eFileInfo && part) {
      this.eFileInfo = eFileInfo;
      if (part.contentLocation) {
        this.location = part.contentLocation;
      } else {
        this.location = "";
        this.cid = part.eParam.cid;
      }
    }
  }

  Object.seal(this);
}
UnMHTSaveParam.prototype = Object.freeze({
  /**
   * 破棄
   */
  destruct: function() {
    this.saveContext = null;

    this.eFileInfo = null;

    this.browser = null;
    this.targetDocument = null;
    this.ownerDocument = null;
    this.useSelection = null;
    this.frameIDs = null;
    if (this.treeWalker) {
      try {
        this.treeWalker.destruct();
      } catch (e) {
      }
      this.treeWalker = null;
    }

    this.content = "";
  },

  /* ==== arINetFileLoaderListener ==== */

  /**
   * 状態変化のイベント
   *
   * @param   {arNetFileLoadContext} context
   *          ファイルロードコンテキスト
   * @param   {arNetFileLoaderStatus} status
   *          状態
   */
  onStatusChange: function(context, status) {
    this.loaderStatus = status;
  },

  /**
   * 取得開始のイベント
   *
   * @param   {arNetFileLoadContext} context
   *          ファイルロードコンテキスト
   * @param   {arNetFileLoaderStatus} status
   *          状態
   * @param   {nsIURI} location
   *          取得中の URI
   * @param   {number} size
   *          ファイルサイズ
   *          不明ならば 0
   */
  onStart: function(context, status, location, size) {
    this.loaderStatus = status;
    this.currentLocation = location.spec;
    this.fileSize = size;

    this.startTime = Date.now();
    this.lastRecvTime = Date.now();
    this.content = "";
  },

  /**
   * 取得終了のイベント
   *
   * @param   {arNetFileLoadContext} context
   *          ファイルロードコンテキスト
   * @param   {arNetFileLoaderStatus} status
   *          状態
   * @param   {?nsIChannel} channel
   *          ネットワークから取得した場合のチャネル
   * @param   {string} contentType
   *          Content-Type
   * @param   {arNetFileError} errorInfo
   *          エラー情報
   */
  onStop: function(context, status, channel, contentType, errorInfo) {
    if (this.status == UnMHTSaveStatus.NG) {
      return;
    }

    if (status == arNetFileLoaderStatus.NG) {
      this.status = UnMHTSaveStatus.NG;
      this.reason = errorInfo.toStringWithBundle(UnMHTStringBundle);
      return;
    }

    if (!this.mimetype) {
      if (contentType) {
        let param
          = arMIMEDecoder.decodeParam(contentType);
        if (param.type && param.subtype) {
          this.mimetype = param.type + "/" + param.subtype;
        }
        if (param.hasParam("charset")) {
          this.charset  = param.getParam("charset");
        }
      }
    }

    if (!this.mimetype ||
        this.mimetype == "application/octet-stream") {
      let ext = arPathUtils.getExtension(this.currentLocation);
      if (ext) {
        this.mimetype = arMIMEUtils.getTypeFromExtension(ext);
      } else if (!this.mimetype) {
        this.mimetype = "application/octet-stream";
      }
    }

    this.lastModified = context._lastModified;

    this.status = UnMHTSaveStatus.OK;

    UnMHTSaveUtils.modifyContent(this, this.currentLocation,
                                 this.saveContext._saveDate, "");
  },

  /**
   * データ到着のイベント
   *
   * @param   {arNetFileLoadContext} context
   *          ファイルロードコンテキスト
   * @param   {arNetFileLoaderStatus} status
   *          状態
   * @param   {string} data
   *          データ
   * @param   {number} count
   *          データの長さ
   */
  onData: function(context, status, data, count) {
    this.content += data;
    this.recvSize += count;
    this.lastRecvTime = Date.now();
  }
});

/**
 * キャッシュ取得のタイムアウト [ms]
 * @type {number}
 */
const CACHE_TIMEOUT = 5 * 1000;

/**
 * ネットワーク取得のタイムアウト [ms]
 * @type {number}
 */
const RECEIVE_TIMEOUT = 15 * 1000;

/**
 * 保存コンテキスト
 *
 * @constructor
 * @param   {nsIDOMWindow} window
 *          実行中のウィンドウ
 * @param   {string} targetFilePath
 *          対象のファイルのパス
 * @param   {string} location
 *          メールを保存する際のルートドキュメントの URI
 * @param   {string} subject
 *          ファイルの subject
 * @param   {nsIDownloadProgressListener} downloadListener
 *          ダウンロードのイベントリスナ
 * @param   {number} saveState
 *          保存状態の数値表現
 * @returns {UnMHTSaveContext}
 *          保存コンテキスト
 */
function UnMHTSaveContext(window, targetFilePath, location, subject,
                          downloadListener, saveState) {
  /**
   * 進捗状況 (0-1)
   * @type {number}
   */
  this.progress = 0;

  /**
   * 実行中のウィンドウ
   * @type {nsIDOMWindow}
   */
  this._window = window;

  /**
   * 対象のファイルのパス
   * @type {string}
   */
  this._targetFilePath = targetFilePath;

  /**
   * ルートドキュメントの URI
   * @type {string}
   */
  this._location = location;

  /**
   * 外部で作成されたダウンロードリスナ
   * @type {nsIWebProgressListener}
   */
  this._downloadListener = downloadListener;

  /**
   * 保存状態
   * @type {UnMHTSaveState}
   */
  this._saveState = saveState;

  /**
   * 保存開始時刻
   * @type {Date}
   */
  this._saveDate = new Date();

  /**
   * トップレベルのパート
   * @type {arMIMEPart}
   */
  this._topPart = UnMHTSaveUtils.createTopPart(subject, this._saveDate);

  /**
   * 開始パート
   * @type {arMIMEPart}
   */
  this._startPart = null;

  /**
   * トップレベルが multipart/mixed か
   * @type {boolean}
   */
  this._isMixed = false;

  /**
   * トップレベルが multipart/alternative か
   * @type {boolean}
   */
  this._isAlternative = false;

  /**
   * 複数のタブを含むかどうか
   * インデックスなしの場合は X-UnMHT-Tabs-List は含めるが
   * X-UnMHT-Tabs は含めない
   * @type {boolean}
   */
  this._isTabs = false;

  /**
   * タブの Content-ID のリスト
   * @type {Array.<string>}
   */
  this._tabs = null;

  /**
   * 収集した URL
   * @type {Set.<string>}
   */
  this._urls = new Set();

  /**
   * 作業中の保存情報
   * @type {Array.<UnMHTSaveParam>}
   */
  this._saveParams = [];

  /**
   * キャンセルされたか
   * @type {boolean}
   */
  this._isCanceled = false;

  /**
   * 現在のファイル
   * @type {UnMHTSaveParam}
   */
  this._currentFile = null;

  /**
   * コンテンツの種類と保存するかの選択のマップ
   *   "save": 保存する
   *   "ignore": 無視する
   *   "ask": ダイアログを表示する
   * @type {Map.<string,string>}
   */
  this._saveContentsAction
    = new Map(["video", "embed"]
              .map(type => [type, UnMHTPref.getPref("save.contents." + type)]));

  /**
   * メールを元のファイルで保存するか
   * @type {boolean}
   */
  this._isMailOriginal = false;

  /**
   * 元のメールのトップパート
   */
  this._mailPart = null;

  /**
   * イベントリスナ
   * @type {IUnMHTSaveListener}
   */
  this._listener = null;

  /**
   * プライベートモードから保存したか
   * @type {boolean}
   */
  this._isPrivate = false;
  try {
    let { PrivateBrowsingUtils } = Cu.import("resource://gre/modules/PrivateBrowsingUtils.jsm", {});
    this._isPrivate = PrivateBrowsingUtils.isWindowPrivate(this._window);
  } catch (e) {
  }

  Object.seal(this);
}
UnMHTSaveContext.prototype = Object.freeze({
  /**
   * 破棄処理
   * start 後は自分で呼ぶ
   */
  destruct: function() {
    if (this._listener) {
      this._listener.onDestruct();
      this._listener = null;
    }

    if (this._saveParams) {
      for (let sParam of this._saveParams) {
        sParam.destruct();
      }
      this._saveParams = null;
    }

    if (this._currentFile) {
      this._currentFile = null;
    }

    if (this._topPart) {
      this._topPart.destruct();
      this._topPart = null;
    }

    if (this._mailPart) {
      this._mailPart.destruct();
      this._mailPart = null;
    }

    this._urls = null;
    this._startPart = null;
    this._downloadListener = null;

    this._window = null;
  },

  /**
   * イベントリスナを設定する
   *
   * @param   {IUnMHTSaveListener} listener
   *          イベントリスナ
   */
  setLisener: function(listener) {
    this._listener = listener;
  },

  /**
   * フィールドを追加する
   *
   * @param   {string} name
   *          フィールド名
   * @param   {string} vlaue
   *          フィールド値
   * @param   {boolean} storeAll
   *          UnMHT が使用するフィールドも無視しない
   */
  addField: function(name, value, storeAll) {
    if (!UnMHTSaveUtils.isFreeField(name) ||
        name.toLowerCase() == "content-type") {
      if (storeAll) {
        name = "X-UnMHT-Original-" + name;
      } else {
        return;
      }
    }
    this._topPart.addField(name, value);
  },

  /**
   * 複数のタブを含むかどうかを指定する
   *
   * @param   {boolean} isTabs
   *          複数のタブを含むかどうか
   */
  setTabs: function(isTabs) {
    this._isTabs = isTabs;
  },

  /**
   * タブのリストを指定する
   *
   * @param   {Array.<string>} tabs
   *          タブの Content-ID のリスト
   */
  setTabList: function(tabs) {
    this._tabs = tabs;
  },

  /**
   * ドキュメントを保存対象に追加する
   *
   * @param   {Object} arg
   *            {
   *              browser:         {nsIXULElement=}
   *                               対象のブラウザ
   *              targetDocument:  {nsIDOMDocument=}
   *                               対象のドキュメント
   *                               (scrapBookData とどちらか必須)
   *              useSelection:    {boolean}
   *                               選択範囲を使用するか
   *              scrapBookData:   {Object=}
   *                               ScrapBook のデータ
   *              saveState:       {number}
   *                               保存状態
   *              location:        {string}
   *                               ドキュメントの URL
   *              isHTML:          {boolean}
   *                               HTML かどうか
   *              isStartPart:     {boolean}
   *                               開始パートか
   *              savedFrom:       {string=}
   *                               保存元
   *                               指定しない場合 location と同じ
   *              needContentID:   {boolean=}
   *                               Content-ID が必要か
   *              window:          {nsIDOMWindow=}
   *                               ドキュメントを持つブラウザを
   *                               実行中のウィンドウ
   *                               メールの添付ファイル取得に使う
   *            }
   * @returns {Promise.<arMIMEPart>}
   * @resolves 対応するパート
   * @rejects UnMHTSaveResult.ERROR_ADD_DOCUMENT
   */
  addDocument: Task.async(function*(arg) {
    if (this._isCanceled) {
      throw UnMHTSaveResult.ERROR_ADD_DOCUMENT;
    }

    let getArg = (name, defValue) => (name in arg) ? arg[name] : defValue;

    let browser         = getArg("browser", null);
    let targetDocument  = getArg("targetDocument", null);
    let useSelection    = getArg("useSelection", null);
    let scrapBookData   = getArg("scrapBookData", null);
    let saveState       = getArg("saveState", UnMHTSaveState.CURRENT);
    let location        = getArg("location", "");
    let isHTML          = getArg("isHTML", false);
    let isStartPart     = getArg("isStartPart", false);
    let savedFrom       = getArg("savedFrom", location);
    let needContentID   = getArg("needContentID", false);
    let window          = getArg("window", null);

    if (UnMHTScheme.isUnMHTURI(location)) {
      let [eFileInfo, currentPart] = lazy.UnMHTCache.getPart(location);
      if (!(eFileInfo && currentPart)) {
        throw UnMHTSaveResult.ERROR_ADD_DOCUMENT;
      }
    }

    let part = null;
    if (scrapBookData) {
      part = yield* this._addScrapBookTask(scrapBookData, isStartPart);
    } else if ((saveState & UnMHTSaveStateMask.ORIGINAL) &&
               UnMHTScheme.isUnMHTURI(location)) {
      part = this._addUnMHTDocument(location, isStartPart, savedFrom,
                                    needContentID);

      this._addNormalDocument(browser,
                              targetDocument, useSelection,
                              saveState,
                              location, false,
                              false, window, true);
    } else if (isHTML) {
      part
        = this._addNormalDocument(browser,
                                  targetDocument, useSelection,
                                  saveState,
                                  location, isStartPart,
                                  false, window, false);
    } else {
      part = this.addFile(browser,
                          targetDocument, location, isStartPart);
    }

    if (!part) {
      throw UnMHTSaveResult.ERROR_ADD_DOCUMENT;
    }

    if (part.sParam) {
      part.sParam.savedFrom = savedFrom;
    }

    return part;
  }),

  /**
   * データドキュメントを保存対象に追加する
   *
   * @param   {string} content
   *          内容
   * @param   {string} location
   *          Content-Location の値
   * @param   {string} charset
   *          Content-Type のうち、文字コード
   * @param   {string} mimetype
   *          Content-Type のうち、MIME タイプ
   * @param   {boolean} isStartPart
   *          開始パートか
   * @returns {?arMIMEPart}
   *          対応するパート
   *          失敗したら null
   */
  addDataDocument: function(content, location, charset, mimetype,
                            isStartPart) {
    let part = new arMIMEPart();
    let sParam = new UnMHTSaveParam(location, this, true);
    part.sParam = sParam;

    sParam.charset = charset;
    sParam.mimetype = mimetype;
    switch (sParam.mimetype) {
      case "text/html": {
        sParam.isHTML = true;
        sParam.isText = true;
        break;
      }
      case "text/css": {
        sParam.isText = true;
        break;
      }
      case "text/javascript": {
        sParam.isText = true;
        break;
      }
    }

    sParam.content = content;

    sParam.status = UnMHTSaveStatus.OK;

    this._saveParams.push(sParam);
    if (isStartPart) {
      this._startPart = part;
      this._topPart.parts.unshift(part);
    } else {
      this._topPart.parts.push(part);
    }

    return part;
  },

  /**
   * ファイルを保存対象に追加する
   *
   * @param   {?nsIXULElement} browser
   *          対象のブラウザ
   * @param   {?nsIDOMDocument} targetDocument
   *          対象のファイルを開いたドキュメント
   *          なければ null
   * @param   {string} location
   *          対象のファイルの URL
   * @param   {boolean} isStartPart
   *          開始パートか
   * @returns {?arMIMEPart}
   *          対応するパート
   *          失敗したら null
   */
  addFile: function(browser, targetDocument, location, isStartPart) {
    if (!UnMHTSaveUtils.checkPolicy(Ci.nsIContentPolicy.TYPE_DOCUMENT,
                                    location, targetDocument)) {
      return null;
    }

    let urlForCheck = UnMHTMailURLUtils.hideMailURL(location, this._location);
    this._urls.add(urlForCheck);

    let part = new arMIMEPart();
    let sParam = new UnMHTSaveParam(location, this, isStartPart);
    sParam.ownerDocument = targetDocument;
    part.sParam = sParam;
    if (targetDocument) {
      try {
        if ("referrer" in targetDocument &&
            targetDocument.referrer) {
          sParam.referrer = targetDocument.referrer;
        }

        sParam.mimetype = targetDocument.contentType;
        sParam.charset = targetDocument.characterSet || "utf-8";
      } catch (e) {
        /* ドキュメントが生きていない */
      }
    }

    this._saveParams.push(sParam);
    if (isStartPart) {
      this._startPart = part;
      this._topPart.parts.unshift(part);
    } else {
      this._topPart.parts.push(part);
    }

    return part;
  },

  /**
   * 保存開始
   *
   * @returns {Promise.<Tuple.<string,Array.<string>>>}
   * @resolves 保存したファイルと保存に失敗したファイルリスト
   * @rejects  {UnMHTSaveResult}
   */
  start: Task.async(function*() {
    for (;;) {
      if (this._isCanceled) {
        if (this._listener) {
          this._listener.onCancel(this);
        }
        this.destruct();

        throw UnMHTSaveResult.CANCEL;
      }

      let [currentFile, currentFileName, currentSize] = this._getCurrent();
      this._currentFile = currentFile;

      let now = Date.now();

      let ok = 0;
      let na_progress = 0;
      let ng = 0;

      for (let sParam of this._saveParams) {
        switch (sParam.status) {
          case UnMHTSaveStatus.FILE_NA: {
            this._startFileLoad(sParam, now);
            break;
          }
          case UnMHTSaveStatus.FILE_GET: {
            if (sParam.fileSize) {
              let progress = sParam.recvSize / sParam.fileSize;
              if (sParam.targetDocument) {
                /* DOM 解析の分の進捗を進める */
                progress = 0.5 + progress * 0.5;
              }
              na_progress += progress;
            }

            if (sParam.loaderStatus == arNetFileLoaderStatus.NA ||
                sParam.loaderStatus & arNetFileLoaderStatusMask.CACHE) {
              if (now > sParam.lastRecvTime + CACHE_TIMEOUT) {
                sParam.startTime = now;
                sParam.lastRecvTime = now;
                sParam.loaderStatus = arNetFileLoaderStatus.NET;

                sParam.loadContext.nextMethod();
              }
            } else if (sParam.loaderStatus == arNetFileLoaderStatus.NET) {
              if (now > sParam.lastRecvTime + RECEIVE_TIMEOUT) {
                sParam.loadContext.cancel();

                sParam.status = UnMHTSaveStatus.NG;
                sParam.reason = TEXT("save.error.timeout");
              }
            }
            break;
          }

          case UnMHTSaveStatus.DOC_CUR_NA: {
            sParam.status = UnMHTSaveStatus.DOC_CUR_FAVICON_START;
            break;
          }
          case UnMHTSaveStatus.DOC_CUR_FAVICON_START: {
            sParam.status = UnMHTSaveStatus.DOC_CUR_FAVICON_WAIT;
            this._startGetFavicon(sParam);
            break;
          }
          case UnMHTSaveStatus.DOC_CUR_FAVICON_WAIT: {
            break;
          }
          case UnMHTSaveStatus.DOC_CUR_SERIAL_START: {
            sParam.status = UnMHTSaveStatus.DOC_CUR_SERIAL_WAIT;
            this._startSerializeDocument(sParam);
            break;
          }
          case UnMHTSaveStatus.DOC_CUR_SERIAL_WAIT: {
            break;
          }
          case UnMHTSaveStatus.DOC_CUR_GATHER_START: {
            sParam.status = UnMHTSaveStatus.DOC_CUR_GATHER_WAIT;
            yield this._startGatherDocument(sParam);
            break;
          }
          case UnMHTSaveStatus.DOC_CUR_GATHER_WAIT: {
            na_progress += 0.5 + sParam.treeWalker.progress * 0.5;
            break;
          }

          case UnMHTSaveStatus.DOC_ORI_NA: {
            sParam.status = UnMHTSaveStatus.DOC_ORI_FAVICON_START;
            break;
          }
          case UnMHTSaveStatus.DOC_ORI_FAVICON_START: {
            sParam.status = UnMHTSaveStatus.DOC_ORI_FAVICON_WAIT;
            this._startGetFavicon(sParam);
            break;
          }
          case UnMHTSaveStatus.DOC_ORI_FAVICON_WAIT: {
            break;
          }
          case UnMHTSaveStatus.DOC_ORI_GATHER_START: {
            sParam.status = UnMHTSaveStatus.DOC_ORI_GATHER_WAIT;
            yield this._startGatherDocument(sParam);
            break;
          }
          case UnMHTSaveStatus.DOC_ORI_GATHER_WAIT: {
            na_progress += sParam.treeWalker.progress * 0.5;
            break;
          }

          case UnMHTSaveStatus.MAIL_CUR_MAIL_START: {
            sParam.status = UnMHTSaveStatus.MAIL_CUR_MAIL_WAIT;
            this._startGetMail(sParam);
            break;
          }
          case UnMHTSaveStatus.MAIL_CUR_MAIL_WAIT: {
            break;
          }

          case UnMHTSaveStatus.MAIL_ORI_MAIL_START: {
            sParam.status = UnMHTSaveStatus.MAIL_ORI_MAIL_WAIT;
            this._startGetMail(sParam);
            break;
          }
          case UnMHTSaveStatus.MAIL_ORI_MAIL_WAIT: {
            break;
          }
          case UnMHTSaveStatus.MAIL_ORI_NA: {
            sParam.status = UnMHTSaveStatus.MAIL_ORI_GATHER_START;
            break;
          }
          case UnMHTSaveStatus.MAIL_ORI_GATHER_START: {
            sParam.status = UnMHTSaveStatus.MAIL_ORI_GATHER_WAIT;
            yield this._startGatherDocument(sParam);
            break;
          }
          case UnMHTSaveStatus.MAIL_ORI_GATHER_WAIT: {
            na_progress += sParam.treeWalker.progress;
            break;
          }

          case UnMHTSaveStatus.REMOTE_NA: {
            sParam.status = UnMHTSaveStatus.REMOTE_GATHER_START;
            break;
          }
          case UnMHTSaveStatus.REMOTE_GATHER_START: {
            sParam.status = UnMHTSaveStatus.REMOTE_GATHER_WAIT;
            yield this._startGatherDocument(sParam);
            break;
          }
          case UnMHTSaveStatus.REMOTE_GATHER_WAIT: {
            na_progress += sParam.treeWalker.progress;
            break;
          }

          case UnMHTSaveStatus.OK: {
            ok += 1;
            break;
          }
          case UnMHTSaveStatus.NG: {
            ng += 1;
            break;
          }
          case UnMHTSaveStatus.IGNORE: {
            ok += 1;
            break;
          }
        }
      }

      let na = this._saveParams.length - (ok + ng);
      if (na > 0) {
        this.progress = (ok + na_progress) / this._saveParams.length;
      } else {
        this.progress = 1;
      }
      if (this._listener) {
        this._listener.onProgress(this, currentFileName, currentSize,
                                  ok, na, ng);
      }

      if (na === 0) {
        break;
      }

      yield arTaskUtils.wait(10);
    }

    if (this._isCanceled) {
      if (this._listener) {
        this._listener.onCancel(this);
      }
      this.destruct();

      throw UnMHTSaveResult.CANCEL;
    }

    if (this._saveParams
        .some(sParam => (sParam.isRequired &&
                         sParam.status == UnMHTSaveStatus.NG))) {
      if (this._listener) {
        this._listener.onError(this, this._createRequiredErrorInfo());
      }
      this.destruct();

      throw UnMHTSaveResult.ERROR_REQUIRED_FAILED;
    }

    this._setPartData(this._topPart);

    this._setTopPartHeader();

    try {
      let data = arMIMEEncoder.encodeMessage(this._topPart);

      yield arFileUtils.createUTF8(this._targetFilePath, data);
    } catch (e) {
      if (this._listener) {
        let errorInfo
          = new UnMHTSaveErrorInfo(UnMHTSaveError.WRITE_FILE,
                                   e,
                                   null);
        this._listener.onError(this, errorInfo);
      }
      this.destruct();

      throw UnMHTSaveResult.ERROR_WRITE_FILE;
    }

    this._registerDownloadHistory();

    let failedList
      = this._saveParams
      .filter(sParam => sParam.status == UnMHTSaveStatus.NG)
      .map(sParam => sParam.currentLocation);

    this.progress = 1;
    if (this._listener) {
      this._listener.onSave(this, failedList);
    }

    this.destruct();

    return [this._targetFilePath, failedList];
  }),

  /**
   * 保存をキャンセル
   */
  cancel: function() {
    this._isCanceled = true;

    if (this._saveParams) {
      for (let sParam of this._saveParams) {
        try {
          sParam.loadContext.cancel();
        } catch (e) {
        }
      }
    }
  },

  /**
   * 現在のファイルをスキップ
   */
  skipFile: function() {
    if (this._currentFile) {
      if (this._currentFile.loadContext) {
        this._currentFile.loadContext.cancel();
      }
      this._currentFile.status = UnMHTSaveStatus.NG;
      this._currentFile.reason = TEXT("save.error.skipped");
    }
  },

  /**
   * 対象のファイルのパス
   * @type {?string}
   */
  get targetFilePath() {
    if (!this._targetFilePath) {
      return null;
    }
    return this._targetFilePath;
  },

  /* ==== private ==== */

  /**
   * ファイルを追加可能かどうかチェックして追加する
   *
   * @param   {string} url
   *          対象の URL
   * @param   {number} type
   *          nsIContentPolicy の対象の種類
   * @param   {nsIDOMDocument} targetDocument
   *          対象のドキュメント
   * @param   {string} referrer
   *          リファラ
   * @param   {string} mimetype
   *          ファイルの MIME-Type
   *          オプショナル
   * @param   {boolean} isRemoteGather
   *          リモートファイルのみ収集するか
   */
  _checkAndPush: function(url, type,
                          targetDocument,
                          referrer,
                          mimetype, isRemoteGather) {
    if (!url) {
      return;
    }

    url = arPathUtils.removeFragment(url);

    if (url.startsWith("data:")) {
      /* data スキームは別途保存する必要が無い */
      return;
    }

    if (url.startsWith("javascript:")) {
      /* 異常 */
      return;
    }

    if (isRemoteGather && url.startsWith("unmht:")) {
      return;
    }

    if (this._isMailOriginal) {
      if (UnMHTMailURLUtils.isMailURL(url)) {
        /* メールを元のファイルで保存する場合は
         * メール内のファイルは別途保存する必要が無い */
        return;
      }
      if (url.startsWith("chrome:")) {
        return;
      }
    }

    if (!mimetype &&
        UnMHTMailURLUtils.isMailURL(url)) {
      mimetype = UnMHTMailURLUtils.getMimeType(url);
    }

    let urlForCheck = UnMHTMailURLUtils.hideMailURL(url, this._location);
    if (!this._urls.has(urlForCheck) &&
        UnMHTSaveUtils.checkPolicy(type, url, targetDocument)) {
      this._urls.add(urlForCheck);

      let part = new arMIMEPart();
      let sParam = new UnMHTSaveParam(url, this, false);
      sParam.ownerDocument = targetDocument;
      part.sParam = sParam;

      sParam.referrer = referrer;

      switch (type) {
        case Ci.nsIContentPolicy.TYPE_STYLESHEET: {
          sParam.isText = true;
          sParam.mimetype = "text/css";
          break;
        }
        case Ci.nsIContentPolicy.TYPE_SCRIPT: {
          sParam.isText = true;
          sParam.mimetype = "text/javascript";
          break;
        }
      }

      if (mimetype) {
        sParam.mimetype = mimetype;
      }

      this._saveParams.push(sParam);
      this._topPart.parts.push(part);
    }
  },

  /**
   * 対象の要素から保存対象のファイルの URL を収集する
   *
   * @param   {nsIXULElement} browser
   *          対象のブラウザ
   * @param   {nsIDOMWindow} targetWindow
   *          対象のウィンドウ
   * @param   {nsIDOMDocument} targetDocument
   *          対象のドキュメント
   * @param   {string} location
   *          ドキュメントの URL
   * @param   {string} savedFrom
   *          ドキュメントの URL
   * @param   {number} saveState
   *          保存状態
   * @param   {nsIDOMElement} node
   *          対象の要素
   * @param   {string} baseDir
   *          ベースディレクトリ
   * @param   {?nsISelection} targetSelection
   *          対象の選択範囲もしくは null
   * @param   {string} referrer
   *          リファラ
   * @param   {boolean} isMail
   *          メールか
   * @param   {boolean} isRemoteGather
   *          リモートファイルのみ収集するか
   * @param   {boolean} saveVideo
   *          VIDEO/AUDIO を保存するか
   * @param   {boolean} saveEmbed
   *          EMBED/OBJECT を保存するか
   */
  _gatherFilesFromNode: function(browser,
                                 targetWindow,
                                 targetDocument, location, savedFrom,
                                 saveState,
                                 node, baseDir,
                                 targetSelection,
                                 referrer, isMail, isRemoteGather,
                                 frameIDs, saveVideo, saveEmbed) {
    if (targetSelection) {
      if (!arDOMUtils.isNodeInSelection(node, targetSelection)) {
        return;
      }
    }

    this._gatherFilesFromNodeCSS(targetWindow,
                                 targetDocument,
                                 node, baseDir,
                                 referrer, isRemoteGather);

    if (node.hasAttribute("background")) {
      this._checkAndPush(arPathUtils.resolve(baseDir,
                                             node.getAttribute("background")),
                         Ci.nsIContentPolicy.TYPE_IMAGE,
                         targetDocument, referrer,
                         null, isRemoteGather);
    }

    let nodeName = node.nodeName.toLowerCase();

    if (node instanceof targetWindow.HTMLImageElement) {
      this._checkAndPush(node.currentSrc, Ci.nsIContentPolicy.TYPE_IMAGE,
                         targetDocument, referrer,
                         null, isRemoteGather);
    } else if (node instanceof targetWindow.HTMLVideoElement) {
      if (saveVideo) {
        this._checkAndPush(node.currentSrc, Ci.nsIContentPolicy.TYPE_MEDIA,
                           targetDocument, referrer,
                           null, isRemoteGather);
        this._checkAndPush(node.poster, Ci.nsIContentPolicy.TYPE_IMAGE,
                           targetDocument, referrer,
                           null, isRemoteGather);
      }
    } else if (node instanceof targetWindow.HTMLAudioElement) {
      if (saveVideo) {
        this._checkAndPush(node.currentSrc, Ci.nsIContentPolicy.TYPE_MEDIA,
                           targetDocument, referrer,
                           null, isRemoteGather);
      }
    } else if (("HTMLTrackElement" in targetDocument &&
                node instanceof targetWindow.HTMLTrackElement) ||
               nodeName == "track") {
      if (saveVideo) {
        this._checkAndPush(arPathUtils.resolve(baseDir,
                                              node.getAttribute("src")),
                           Ci.nsIContentPolicy.TYPE_MEDIA,
                           targetDocument, referrer,
                           null, isRemoteGather);
      }
    } else if (node instanceof targetWindow.HTMLSourceElement) {
      if (saveVideo) {
        this._checkAndPush(node.currentSrc, Ci.nsIContentPolicy.TYPE_MEDIA,
                           targetDocument, referrer,
                           null, isRemoteGather);
      }
    } else if (node instanceof targetWindow.HTMLLinkElement &&
               node.rel) {
      if (/\bicon\b/i.test(node.rel)) {
        this._checkAndPush(node.href, Ci.nsIContentPolicy.TYPE_IMAGE,
                           targetDocument, referrer,
                           null, isRemoteGather);
      }
    } else if (node instanceof targetWindow.HTMLInputElement &&
               node.type.toLowerCase() == "image") {
      this._checkAndPush(node.src, Ci.nsIContentPolicy.TYPE_IMAGE,
                         targetDocument, referrer,
                         null, isRemoteGather);
    } else if (node instanceof targetWindow.HTMLScriptElement) {
      if (saveState != UnMHTSaveState.CURRENT_NOJS) {
        this._checkAndPush(node.src, Ci.nsIContentPolicy.TYPE_SCRIPT,
                           targetDocument, referrer,
                           null, isRemoteGather);
      }
    } else if (node instanceof targetWindow.HTMLEmbedElement) {
      if (saveEmbed) {
        this._checkAndPush(arPathUtils.resolve(baseDir, node.src),
                           Ci.nsIContentPolicy.TYPE_OBJECT,
                           targetDocument, referrer,
                           node.getAttribute("type"), isRemoteGather);
      }
    } else if (node instanceof targetWindow.HTMLObjectElement) {
      if (saveEmbed) {
        this._checkAndPush(arPathUtils.resolve(baseDir, node.data),
                           Ci.nsIContentPolicy.TYPE_OBJECT,
                           targetDocument, referrer,
                           node.getAttribute("type"), isRemoteGather);
      }
    } else if (node instanceof targetWindow.HTMLFrameElement ||
               node instanceof targetWindow.HTMLIFrameElement) {
      let cid = frameIDs.has(node) ? frameIDs.get(node) : null;

      let frameLocation = node.contentDocument.documentURI;

      if (frameLocation.startsWith("javascript:") ||
          frameLocation.startsWith("data:") ||
          frameLocation == "about:srcdoc") {
        frameLocation = location;
      }

      this._addNormalDocument(browser,
                              node.contentDocument, null,
                              saveState,
                              frameLocation, false,
                              node instanceof targetWindow.HTMLIFrameElement,
                              null, isRemoteGather, cid);
    }
  },


  /**
   * 対象の要素のスタイルから保存対象のファイルの URL を収集する
   *
   * @param   {nsIDOMWindow} targetWindow
   *          対象のウィンドウ
   * @param   {nsIDOMDocument} targetDocument
   *          対象のドキュメント
   * @param   {nsIDOMElement} node
   *          対象の要素
   * @param   {string} baseDir
   *          ベースディレクトリ
   * @param   {string} referrer
   *          リファラ
   * @param   {boolean} isRemoteGather
   *          リモートファイルのみ収集するか
   */
  _gatherFilesFromNodeCSS: function(targetWindow,
                                    targetDocument,
                                    node, baseDir,
                                    referrer, isRemoteGather) {
    let pseudos = [
      "",
      ":before",
      ":after",
      ":first-letter",
      ":first-line"
      /*
       * :selection は色のみなので無視
       */
    ];

    let styleTypes = new Map();
    styleTypes.set("background-image", Ci.nsIContentPolicy.TYPE_IMAGE);
    styleTypes.set("list-style-image", Ci.nsIContentPolicy.TYPE_IMAGE);
    styleTypes.set("border-image-source", Ci.nsIContentPolicy.TYPE_IMAGE);
    styleTypes.set("cursor", Ci.nsIContentPolicy.TYPE_IMAGE);
    styleTypes.set("content", Ci.nsIContentPolicy.TYPE_IMAGE);
    styleTypes.set("filter", Ci.nsIContentPolicy.TYPE_OTHER);
    styleTypes.set("clip-path", Ci.nsIContentPolicy.TYPE_OTHER);
    styleTypes.set("mask", Ci.nsIContentPolicy.TYPE_OTHER);

    for (let pseudo of pseudos) {
      let definedStyles
        = UnMHTSaveUtils.getDefinedStyles(node, pseudo, styleTypes);
      if (!definedStyles) {
        continue;
      }

      let style = null;
      try {
        style = targetDocument.defaultView.getComputedStyle(node, pseudo);
      } catch (e) {
        continue;
      }

      if (!style) {
        continue;
      }

      for (let name of definedStyles) {
        let type = styleTypes.get(name);

        let value = null;
        try {
          value = style.getPropertyCSSValue(name);
        } catch (e) {
          continue;
        }
        this._checkCSS(targetWindow,
                       targetDocument,
                       baseDir,
                       referrer, isRemoteGather,
                       value, type);
      }
    }
  },

  /**
   * 対象の要素のスタイルから保存対象のファイルの URL を収集する
   *
   * @param   {nsIDOMWindow} targetWindow
   *          対象のウィンドウ
   * @param   {nsIDOMDocument} targetDocument
   *          対象のドキュメント
   * @param   {string} baseDir
   *          ベースディレクトリ
   * @param   {string} referrer
   *          リファラ
   * @param   {boolean} isRemoteGather
   *          リモートファイルのみ収集するか
   * @param   {nsIDOMCSSValue} value
   *          CSS の値
   * @param   {number} type
   *          ファイルの種類
   *            Ci.nsIContentPolicy.TYPE_*
   */
  _checkCSS: function(targetWindow,
                      targetDocument,
                      baseDir,
                      referrer, isRemoteGather,
                      value, type) {
    if (!value) {
      return;
    }

    if (value.cssValueType == targetWindow.CSSValue.CSS_PRIMITIVE_VALUE) {
      if (value.primitiveType == targetWindow.CSSPrimitiveValue.CSS_URI) {
        let url = arPathUtils.resolve(baseDir, value.getStringValue());
        this._checkAndPush(url, type,
                           targetDocument, referrer,
                           null, isRemoteGather);
      } else if (value.primitiveType == targetWindow.CSSPrimitiveValue.CSS_STRING) {
        value.getStringValue()
          .replace(/url\s*\(\s*(?:\"((?:\\\"|[^\"])+)\"|\'((?:\\\'|[^\'])+)\'|([^\"\';\(\)]+))/ig,
                   (matched, dqpath, qpath, path) => {
                     if (dqpath) {
                       path = dqpath.replace(/\\\"/g, "\"");
                     } else if (qpath) {
                       path = qpath.replace(/\\\'/g, "\'");
                     }

                     path = path.replace(/ +$/, "");

                     this._checkAndPush(arPathUtils.resolve(baseDir, path),
                                        type,
                                        targetDocument, referrer,
                                        null, isRemoteGather);
                   });
      }
    } else if (value.cssValueType == targetWindow.CSSValue.CSS_VALUE_LIST) {
      for (let item of Array.from(value)) {
        this._checkCSS(targetWindow,
                       targetDocument,
                       baseDir,
                       referrer, isRemoteGather,
                       item, type);
      }
    }
  },

  /**
   * コンテンツを保存するかどうか調べる
   *
   * @param   {string} type
   *          種類
   * @yields  {Promise.<*>}
   * @returns {boolean}
   *          保存するか
   */
  _checkSaveContentsTask: function*(type) {
    let action = this._saveContentsAction.get(type);
    if (arAppInfo.isFennec && action == "ask") {
      action = "save";
    }
    if (action == "ask") {
      action = yield UnMHTSaveUtils.showSaveContentsConfirm(this._window,
                                                            type);
      this._saveContentsAction.set(type, action);
    }

    return action == "save";
  },

  /**
   * 通常の HTML ドキュメントを保存対象に追加する
   *
   * @param   {?nsIXULElement} browser
   *          対象のブラウザ
   * @param   {nsIDOMDocument} targetDocument
   *          対象のドキュメント
   * @param   {boolean} useSelection
   *          選択範囲を使用するか
   * @param   {number} saveState
   *          保存状態
   * @param   {string} location
   *          ドキュメントの URL
   * @param   {boolean} isStartPart
   *          開始パートか
   * @param   {boolean} isIFrame
   *          インラインフレームか
   * @param   {?nsIDOMWindow} window
   *          ドキュメントを持つブラウザを実行中のウィンドウ
   *          メールの添付ファイル取得に使う
   * @param   {boolean} isRemoteGather
   *          リモートファイルのみ収集するか
   * @returns {?arMIMEPart}
   *          メインのパート
   *          失敗したら null
   */
  _addNormalDocument: function(browser,
                               targetDocument, useSelection,
                               saveState,
                               location, isStartPart,
                               isIFrame,
                               window, isRemoteGather,
                               cid) {
    if (!isRemoteGather) {
      let urlForCheck = UnMHTMailURLUtils.hideMailURL(location, this._location);
      this._urls.add(urlForCheck);
    }

    let part = null;
    if (!isRemoteGather) {
      part = new arMIMEPart();
    }
    let sParam = new UnMHTSaveParam(location, this, true);
    sParam.browser = browser;
    sParam.ownerDocument = targetDocument;
    if (part) {
      part.sParam = sParam;
    }

    try {
      if ("referrer" in targetDocument &&
          targetDocument.referrer) {
        sParam.referrer = targetDocument.referrer;
      }
    } catch (e) {
      if (part) {
        part.destruct();
      }
      sParam.destruct();
      return null;
    }

    sParam.isIFrame = isIFrame;
    if (cid) {
      sParam.cid = cid;
      sParam.storeContentID = true;
    }
    sParam.isHTML = true;
    sParam.isXHTML = UnMHTSaveUtils.isXHTML(targetDocument);
    sParam.isText = true;
    sParam.isMail = UnMHTMailURLUtils.isMailURL(location);
    sParam.mimetype = "text/html";
    try {
      sParam.charset = targetDocument.characterSet || "utf-8";
    } catch (e) {
      sParam.charset = "utf-8";
    }

    if (isStartPart && sParam.eFileInfo) {
      /* UnMHT の展開した結果であり
       * 開始パートならば展開結果のヘッダを継承する */
      this._addAllFields(sParam.eFileInfo.topPart, false);
    }

    if (!this._getBaseDir(targetDocument, location, sParam)) {
      if (part) {
        part.destruct();
      }
      sParam.destruct();
      return null;
    }

    sParam.targetDocument = targetDocument;
    sParam.useSelection = useSelection;
    sParam.saveState = saveState;
    sParam.frameIDs = new Map();

    if ((saveState & UnMHTSaveStateMask.CURRENT) ||
        useSelection) {
      sParam.status = UnMHTSaveStatus.DOC_CUR_NA;
    } else if (isRemoteGather) {
      sParam.status = UnMHTSaveStatus.REMOTE_NA;
    } else {
      sParam.status = UnMHTSaveStatus.DOC_ORI_NA;
    }
    this._saveParams.push(sParam);

    if (isRemoteGather) {
      return part;
    }

    if (isStartPart) {
      this._startPart = part;
      this._topPart.parts.unshift(part);
    } else {
      this._topPart.parts.push(part);
    }

    if (sParam.isMail && window) {
      if (sParam.status == UnMHTSaveStatus.DOC_ORI_NA) {
        this._isMailOriginal = true;
      } else if (sParam.status == UnMHTSaveStatus.DOC_CUR_NA) {
        if ("currentAttachments" in window &&
            window.currentAttachments) {
          sParam.attachments
            = window.currentAttachments.map(attachment => {
              if ((("hasFile" in attachment && attachment.hasFile) ||
                   !("hasFile" in attachment)) &&
                  !("isDeleted" in attachment && attachment.isDeleted) &&
                  !("isExternalAttachment" in attachment && attachment.isExternalAttachment) &&
                  attachment.url) {
                let filePart = this.addFile(null, null, attachment.url, false);
                filePart.sParam.mimetype = attachment.contentType;
                let name = "attachment";
                if ("displayName" in attachment && attachment.displayName) {
                  name = attachment.displayName;
                }
                if ("name" in attachment && attachment.name) {
                  name = attachment.name;
                }
                return [name, attachment.url];
              }
              return null;
            })
            .filter(attachment => attachment);
        }
      }

      if ("gFolderDisplay" in window &&
          "selectedMessageUris" in window.gFolderDisplay &&
          window.gFolderDisplay.selectedMessageUris.length == 1) {
        let messageURL
          = lazy.MsgMailSession.ConvertMsgURIToMsgURL(window.gFolderDisplay.selectedMessageUris[0], window.msgWindow);

        let uri;
        try {
          uri = Services.io.newURI(messageURL, null, null);
        } catch (e) {
        }

        if (uri) {
          if (sParam.status == UnMHTSaveStatus.DOC_CUR_NA) {
            sParam.status = UnMHTSaveStatus.MAIL_CUR_MAIL_START;
          } else if (sParam.status == UnMHTSaveStatus.DOC_ORI_NA) {
            sParam.status = UnMHTSaveStatus.MAIL_ORI_MAIL_START;
          }

          sParam.messageURI = uri;
        }

        if ("gExpandedHeaderList" in window &&
            "gExpandedHeaderView" in window &&
            "currentHeaderData" in window &&
            UnMHTPref.getPref("save.header.mail.show")) {
          this._getMailHeader(window, sParam);
        }
      }
    }

    return part;
  },

  /**
   * メールヘッダーの情報を取得する
   *
   * @param   {?nsIDOMWindow} window
   *          ドキュメントを持つブラウザを実行中のウィンドウ
   *          メールの添付ファイル取得に使う
   * @param   {UnMHTSaveParam} sParam
   *          保存情報
   */
  _getMailHeader: function(window, sParam) {
    try {
      let display = [];

      let { MailServices } = Cu.import("resource:///modules/mailServices.js", {});

      let used = new Set();

      let emailFormatter = value => {
        if (!value) {
          return "";
        }

        let addresses = {};
        let fullNames = {};
        let names = {};

        let count = MailServices.headerParser
          .parseHeadersWithArray(value, addresses, names, fullNames);

        let values = [];
        for (let i = 0; i < count; i += 1) {
          values.push(fullNames.value[i] || names.value[i]);
        }

        return values.join(", ");
      };
      let newsgroupsFormatter = value => {
        return value.split(",").join(", ");
      };
      let messageIdFormatter = value => {
        return value.split(/\s+/).join(", ");
      };

      let addToDisplay = lowerHeaderName => {
        if (used.has(lowerHeaderName)) {
          return false;
        }

        if (!(lowerHeaderName in window.gExpandedHeaderView)) {
          return false;
        }
        let view = window.gExpandedHeaderView[lowerHeaderName];
        if (!view || !view.valid) {
          return false;
        }

        if (!(lowerHeaderName in window.currentHeaderData)) {
          return false;
        }
        let data = window.currentHeaderData[lowerHeaderName];
        if (!data) {
          return false;
        }

        let formatters = {
          "from": emailFormatter,
          "reply-to": emailFormatter,
          "to": emailFormatter,
          "cc": emailFormatter,
          "bcc": emailFormatter,

          "newsgroups": newsgroupsFormatter,
          "references": messageIdFormatter,
          "followup-to": newsgroupsFormatter,
          "message-id": messageIdFormatter,
          "in-reply-to": messageIdFormatter,
          "sender": messageIdFormatter
        };

        let value = "";

        if (lowerHeaderName in formatters) {
          value = formatters[lowerHeaderName](data.headerValue);
        } else {
          value = data.headerValue;
        }

        display.push([data.headerName, value]);

        used.add(lowerHeaderName);

        return true;
      };

      for (let header of window.gExpandedHeaderList) {
        addToDisplay(header.name);
      }

      if (addToDisplay("x-mozilla-localizeddate")) {
        used.add("date");
      }

      for (let lowerHeaderName of Object.keys(window.currentHeaderData)) {
        addToDisplay(lowerHeaderName);
      }

      sParam.mailHeaderDisplay = display;
    } catch (e) {
      Cu.reportError(e + "\n" + e.stack);
    }
  },

  /**
   * ベースディレクトリを取得する
   *
   * @param   {nsIDOMDocument} targetDocument
   *          対象のドキュメント
   * @param   {string} location
   *          ドキュメントの URL
   * @param   {UnMHTSaveParam} sParam
   *          対象の保存情報
   * @returns {boolean}
   *          成功したか
   */
  _getBaseDir: function(targetDocument, location, sParam) {
    let hasBaseElement = false;
    let baseDir
      = arPathUtils.getBaseDir(UnMHTSaveUtils.getMHTLocation(location));
    try {
      for (let node of Array.from(targetDocument.getElementsByTagName("base"))) {
        hasBaseElement = true;
        if (node.hasAttribute("href") &&
            node.getAttribute("href")) {
          let baseURI;
          try {
            let href = UnMHTSaveUtils.getMHTLocation(node.getAttribute("href"));
            baseURI = Services.io.newURI(href, null, null);
          } catch (e) {
            continue;
          }
          baseDir = arPathUtils.resolve(baseDir, baseURI.resolve("./"));
        }
      }
    } catch (e) {
      return false;
    }

    sParam.baseDir = baseDir;
    sParam.hasBaseElement = hasBaseElement;

    return true;
  },

  /**
   * UnMHT で展開したドキュメントを保存対象に追加する
   *
   * @param   {string} location
   *          対象の URI
   * @param   {boolean} isStartPart
   *          開始パートか
   * @param   {string} savedFrom
   *          保存元のアドレス
   * @param   {boolean} needContentID
   *          Content-ID が必要か
   * @returns {?arMIMEPart}
   *          メインのパート
   *          失敗したら null
   */
  _addUnMHTDocument: function(location, isStartPart, savedFrom,
                              needContentID) {
    let [eFileInfo, currentPart] = lazy.UnMHTCache.getPart(location);
    if (!(eFileInfo && currentPart)) {
      return null;
    }

    if (isStartPart && currentPart == eFileInfo.startPart) {
      this._addAllFields(eFileInfo.topPart, false);
    }

    let createSaveParam = part => {
      if (part.isMultipart) {
        for (let p of part.parts) {
          createSaveParam(p);
        }
        return;
      }

      if (part.mimetype == "text/html") {
        let sParam = new UnMHTSaveParam(location, this, true);

        sParam.isHTML = true;
        sParam.content = part.body;

        UnMHTSaveUtils.modifyContent(sParam, savedFrom, this._saveDate,
                                     "");

        part.body = sParam.content;

        sParam.destruct();
      }
    };

    let findPart = {
      from: currentPart,
      to: null
    };
    let part = eFileInfo.topPart.clone(findPart);
    let mainPart = findPart.to;

    createSaveParam(part);

    if (needContentID && !mainPart.contentID) {
      mainPart.contentID = currentPart.eParam.cid;
      mainPart.addField("Content-ID", mainPart.contentID);
    }

    if (isStartPart) {
      if (part.isMultipart) {
        this._startPart = part.findStartPart();
        if (this._startPart.mimetype == "multipart/mixed") {
          this._startPart = null;
          this._isMixed = true;
        } else if (this._startPart.mimetype == "multipart/alternative") {
          this._startPart = null;
          this._isAlternative = true;
        }

        for (let p of part.parts) {
          this._topPart.parts.push(p);
        }
      } else {
        this._startPart = part;
        this._topPart.parts.unshift(part);
      }
    } else {
      for (let name of part.fieldNames) {
        if (!UnMHTSaveUtils.isFreeField(name)) {
          part.removeField(name);
        }
      }

      this._topPart.parts.push(part);
    }

    return mainPart;
  },

  /**
   * ScrapBook のデータを保存対象に追加する
   *
   * @param   {Object} scrapBookData
   *          ScrapBook のデータ
   * @param   {boolean} isStartPart
   *          開始パートか
   * @yields  {Promise.<*>}
   * @returns {?arMIMEPart}
   *          メインのパート
   *          失敗したら null
   */
  _addScrapBookTask: function*(scrapBookData, isStartPart) {
    let mainPart = null;

    let resIter = new OS.File.DirectoryIterator(scrapBookData["dir"]);
    let resEntries = yield resIter.nextBatch();

    for (let entry of resEntries) {
      if (entry.isDir) {
        continue;
      }

      if (entry.name == "index.dat") {
        continue;
      }

      let isIndex = false;
      if (entry.name == "index.html") {
        isIndex = true;
      }

      let part = new arMIMEPart();
      let sParam = new UnMHTSaveParam(entry.name, this, true);
      part.sParam = sParam;

      sParam.referrer = "";

      if (isIndex) {
        sParam.location = scrapBookData["source"];
        sParam.mimetype = "text/html";
        sParam.charset = scrapBookData["charset"];
      } else {
        sParam.location
          = arPathUtils.resolve(arPathUtils.getBaseDir(scrapBookData["source"]),
                                sParam.location);
        let ext = arPathUtils.getExtension(sParam.location);
        if (ext) {
          sParam.mimetype = arMIMEUtils.getTypeFromExtension(ext);
        }
      }
      sParam.savedFrom = sParam.location;
      sParam.currentLocation = sParam.location;

      sParam.content = yield arFileUtils.readBinary(entry.path);

      UnMHTSaveUtils.modifyContent(sParam, sParam.savedFrom, this._saveDate,
                                   "");

      sParam.status = UnMHTSaveStatus.OK;

      this._saveParams.push(sParam);
      if (isIndex) {
        mainPart = part;
        this._topPart.parts.unshift(part);
        if (isStartPart) {
          this._startPart = part;
        }
      } else {
        this._topPart.parts.push(part);
      }
    }

    return mainPart;
  },

  /**
   * 現在取得中のファイル情報を返す
   *
   * @returns {Tuple.<UnMHTSaveParam,string,string>}
   *          保存情報とファイルの URI と進捗状況
   */
  _getCurrent: function() {
    for (let sParam of this._saveParams) {
      switch (sParam.status) {
        case UnMHTSaveStatus.DOC_CUR_GATHER_WAIT:
        case UnMHTSaveStatus.DOC_ORI_GATHER_WAIT:
        case UnMHTSaveStatus.MAIL_ORI_GATHER_WAIT:
        case UnMHTSaveStatus.REMOTE_GATHER_WAIT: {
          return [sParam,
                  sParam.currentLocation,
                  "scanning DOM nodes: "
                  + Math.floor(100 * sParam.treeWalker.progress) + "%"];
        }
        case UnMHTSaveStatus.OK:
        case UnMHTSaveStatus.NG: {
          break;
        }

        case UnMHTSaveStatus.FILE_GET: {
          if (sParam.fileSize) {
            return [sParam,
                    sParam.currentLocation,
                    sParam.recvSize + " bytes / " + sParam.fileSize + " bytes"];
          }
        }
        /* falls through */
        default: {
          return [sParam,
                  sParam.currentLocation,
                  sParam.recvSize + " bytes / "
                  + (sParam.fileSize || "unknown") + " bytes"];
        }
      }
    }

    return [null, "", ""];
  },

  /**
   * 必須ファイル取得失敗のエラー情報を作成する
   *
   * @returns {UnMHTSaveErrorInfo}
   *          エラー情報
   */
  _createRequiredErrorInfo: function() {
    let errorList
      = this._saveParams
      .filter(sParam => (sParam.isRequired &&
                         sParam.status == UnMHTSaveStatus.NG))
      .map(sParam => [sParam.savedFrom, sParam.reason]);
    return new UnMHTSaveErrorInfo(UnMHTSaveError.REQUIRED_FAILED,
                                  null,
                                  errorList);
  },

  /**
   * トップパートのヘッダを設定する
   */
  _setTopPartHeader: function() {
    if (this._startPart &&
        (this._startPart.mimetype ||
         (this._startPart.sParam &&
          this._startPart.sParam.mimetype))) {
      if (this._startPart.mimetype) {
        this._topPart.startMimetype = this._startPart.mimetype;
      } else {
        this._topPart.startMimetype = this._startPart.sParam.mimetype;
      }

      let params = new arMIMEParams();
      params.type = "multipart";
      params.subtype = "related";
      params.addParam("boundary", this._topPart.boundary);
      params.addParam("type", this._topPart.startMimetype);
      this._topPart.setContentType(params);
    } else if (this._isMixed) {
      this._topPart.startMimetype = "";

      let params = new arMIMEParams();
      params.type = "multipart";
      params.subtype = "mixed";
      params.addParam("boundary", this._topPart.boundary);
      this._topPart.setContentType(params);
    } else if (this._isAlternative) {
      this._topPart.startMimetype = "";

      let params = new arMIMEParams();
      params.type = "multipart";
      params.subtype = "alternative";
      params.addParam("boundary", this._topPart.boundary);
      this._topPart.setContentType(params);
    } else {
      /* **** FIXME: この分岐は本当は来てはいけない **** */

      if (this._startPart) {
        Cu.reportError("FIXME: UnMHT: MIME-Type for startPart is not found for multipart/related.");
      } else {
        Cu.reportError("FIXME: UnMHT: startPart is not found for multipart/related.");
      }

      let params = new arMIMEParams();
      params.type = "multipart";
      params.subtype = "related";
      params.addParam("boundary", this._topPart.boundary);
      this._topPart.setContentType(params);
    }

    this._topPart.removeField("X-UnMHT-Tabs");
    if (this._isTabs) {
      this._topPart.addField("X-UnMHT-Tabs", "true");
    }
    this._topPart.removeField("X-UnMHT-Tabs-List");
    if (this._tabs) {
      this._topPart.addField("X-UnMHT-Tabs-List", this._tabs.map(cid => "<" + cid + ">").join(","));
    }
    this._topPart.removeField("X-UnMHT-Save-State");
    switch (this._saveState) {
      case UnMHTSaveState.CURRENT: {
        this._topPart.addField("X-UnMHT-Save-State", "Current-State");
        break;
      }
      case UnMHTSaveState.CURRENT_NOJS: {
        this._topPart.addField("X-UnMHT-Save-State", "Current-State-Without-Script");
        break;
      }
      case UnMHTSaveState.ORIGINAL: {
        this._topPart.addField("X-UnMHT-Save-State", "Original-File");
        break;
      }
    }
  },

  /**
   * 各パートのヘッダとボディを設定する
   *
   * @param   {arMIMEPart} part
   *          対象のパート
   */
  _setPartData: function(part) {
    if (part.isMultipart) {
      part.parts = part.parts.filter(p => this._setPartData(p));
      return true;
    }

    let sParam = part.sParam;
    if (!sParam) {
      return true;
    }

    if (sParam.status == UnMHTSaveStatus.NG ||
        sParam.status == UnMHTSaveStatus.IGNORE) {
      part.destruct();
      return false;
    }

    let isUTF16 = false;
    if (sParam.charset && sParam.charset.match(/^UTF-16/i)) {
      isUTF16 = true;
    }

    part.contentTransferEncoding
      = (sParam.isText && !isUTF16) ? "quoted-printable" : "base64";
    part.charset = sParam.charset;
    part.mimetype = sParam.mimetype;
    if (UnMHTPref.getPref("save.header.last_modified") &&
        !part.hasField("X-UnMHT-Last-Modified")) {
      try {
        let lastModified = new Date(sParam.lastModified);
        if (!Number.isNaN(lastModified.getFullYear())) {
          part.addField("X-UnMHT-Last-Modified",
                        arMIMEEncoder.encodeDate(lastModified));
        }
      } catch (e) {
        /* フォーマットが不明 */
      }
    }

    let location = sParam.location;
    if (sParam.cid) {
      part.contentID = sParam.cid;
    }

    let storeContentID = sParam.storeContentID;
    if (UnMHTMailURLUtils.isMailURL(location)) {
      if (this._mailPart) {
        let m = location.match(/^.+\?part=([^&]+).*/);
        if (m) {
          let partID = m[1];
          let sourcePart
            = partID.split(".")
            .map(i => Number.parseInt(i, 10))
            .slice(1)
            .reduce((p, i) => {
              if (!p || !p.parts || i - 1 >= p.parts.length) {
                return null;
              }

              return p.parts[i - 1];
            }, this._mailPart);
          if (sourcePart &&
              sourcePart.contentID) {
            storeContentID = true;
            part.contentID = sourcePart.contentID;
          }
        }
      }

      location = UnMHTMailURLUtils.hideMailURL(location, this._location);
    }
    part.contentLocation = location;

    if (part.contentID && part.contentLocation == "about:blank") {
      part.contentLocation = "";
    }

    part.body = sParam.content;

    let params = new arMIMEParams(part.mimetype);
    if (part.charset) {
      params.addParam("charset", part.charset);
    }
    part.setContentType(params);
    part.addField("Content-Transfer-Encoding",
                  part.contentTransferEncoding);

    if (!part.contentLocation ||
        storeContentID) {
      part.addField("Content-ID",
                    "<" + part.contentID + ">");
    }

    if (part.contentLocation) {
      part.addField("Content-Location",
                    part.contentLocation);
    }

    return true;
  },

  /**
   * フィールドを追加する
   *
   * @param   {arMIMEPart} part
   *          対象のパート
   * @param   {boolean} storeAll
   *          UnMHT が使用するフィールドも無視しない
   */
  _addAllFields: function(part, storeAll) {
    for (let [name, value] of part.fieldList) {
      this.addField(name, value, storeAll);
    }
  },

  /**
   * ダウンロード履歴に登録する
   */
  _registerDownloadHistory: function() {
    if (this._downloadListener) {
      this._notifyDownloadListener(this._downloadListener).catch(e => {
        Cu.reportError(e + "\n" + e.stack);
      });
      return;
    }

    if (!UnMHTPref.getPref("save.history.register")) {
      return;
    }

    try {
      let source = Services.io.newURI(this._location, null, null);
      let targetFile = arFileUtils.getFile(this._targetFilePath);
      let target = Services.io.newFileURI(targetFile);
      let cancelable = {
        QueryInterface: XPCOMUtils.generateQI([Ci.nsICancelable]),
        cancel: function(reason) {}
      };

      let transfer
        = Cc["@mozilla.org/transfer;1"].createInstance(Ci.nsITransfer);
      /* nsITransfer.init の aMIMEInfo (第 4 引数) に
       * preferredAction が Ci.nsIMIMEInfo.saveToDisk でない
       * nsIMIMEInfo を実装したインスタンスを渡した場合
       * 登録されたハンドラによってダウンロード完了時にファイルが開かれる
       * また、ハンドラに渡す一時ファイル扱いになり
       * 以下の場合にブラウザ終了時に削除される
       *  1. プライベートモード
       *  2. browser.helperApps.deleteTempFileOnExit が true
       * なので、必ず null を渡す
       */
      transfer.init(source, target, "", null, null, null, cancelable,
                    this._isPrivate);

      this._notifyDownloadListener(transfer);
    } catch (e) {
      Cu.reportError(e + "\n" + e.stack);
    }
  },

  /**
   * ダウンロードのイベントリスナにイベントを通知する
   *
   * @param   {nsIDownloadProgressListener} listener
   *          ダウンロードのイベントリスナ
   */
  _notifyDownloadListener: Task.async(function*(listener) {
    let request = null;
    /* 保存した直後にファイルが消える事もありうるので
     * 消えている場合を想定する */
    let fileSize = 0;
    let exists = yield OS.File.exists(this._targetFilePath);
    if (exists) {
      let stat = yield OS.File.stat(this._targetFilePath);
      fileSize = stat.size;
    }

    let maxFileSize = fileSize;

    listener.onStateChange(null, request,
                           Ci.nsIWebProgressListener.STATE_START
                           | Ci.nsIWebProgressListener.STATE_IS_NETWORK,
                           Cr.NS_OK);
    listener.onProgressChange(null, request, fileSize, maxFileSize,
                              fileSize, maxFileSize);
    listener.onStateChange(null, request,
                           Ci.nsIWebProgressListener.STATE_STOP
                           | Ci.nsIWebProgressListener.STATE_IS_NETWORK,
                           Cr.NS_OK);
  }),

  /**
   * ドキュメントをシリアライズする
   *
   * @param   {UnMHTSaveParam} sParam
   *          対象の保存情報
   */
  _startSerializeDocument: function(sParam) {
    setTimeout(() => {
      try {
      sParam.content = "";

      if (sParam.targetDocument.doctype) {
        sParam.content = UnMHTSaveUtils.serializeDocType(sParam.targetDocument);
      }

      let tmpNodes = [];
      let ignoreNodes = [];

      let nodes = sParam.targetDocument.getElementsByTagName("head");
      if (nodes.length > 0) {
        let head = nodes[0];

        if (sParam.faviconData && !UnMHTSaveUtils.hasFaviconNode(head)) {
          let node = UnMHTSaveUtils.makeFaviconNode(sParam.targetDocument,
                                                    sParam.faviconData,
                                                    sParam.faviconType);
          head.insertBefore(node, head.firstChild);
          tmpNodes.push(node);
        }

        if (sParam.targetDocument.characterSet) {
          let [node, others]
            = UnMHTSaveUtils.makeCharsetNode(sParam.targetDocument, head);
          head.insertBefore(node, head.firstChild);
          tmpNodes.push(node);
          ignoreNodes = ignoreNodes.concat(others);
        }

        if (!sParam.hasBaseElement &&
            sParam.baseDir &&
            sParam.baseDir != "about:blank") {
          let node = UnMHTSaveUtils.makeBaseNode(sParam.targetDocument,
                                                 sParam.baseDir);
          head.appendChild(node);
          tmpNodes.push(node);
        }
      }

      let markerID = "marker." + arPseudoID.gen() + "@unmht.org";
      let params = {
        noScript: sParam.saveState == UnMHTSaveState.CURRENT_NOJS,
        isMail: sParam.isMail,
        saveCanvas: sParam.saveState == UnMHTSaveState.CURRENT_NOJS,
        ignoreNodes: ignoreNodes,
        frameIDs: sParam.frameIDs,
        markerID: markerID,
        isXML: sParam.isXHTML
      };
      sParam.content
        += UnMHTDOMConverter.toPlainText(this._window, sParam.location,
                                         sParam.targetDocument,
                                         sParam.useSelection, params);

      for (let node of tmpNodes) {
        node.remove();
      }

      if (sParam.mailHeaderDisplay) {
        let header = UnMHTSaveUtils.makeMailHeaderDisplay(sParam.mailHeaderDisplay,
                                                          sParam.isXHTML);
        sParam.content
          = UnMHTSaveUtils.modifyBody(sParam.content,
                                      header + "<div style=\"position: relative\">",
                                      "</div>",
                                      markerID);
      }

      if (sParam.attachments) {
        let box = UnMHTSaveUtils.makeAttachmentsBox(sParam.attachments,
                                                    sParam.location);

        sParam.content
          = UnMHTSaveUtils.modifyBody(sParam.content,
                                      "",
                                      box,
                                      markerID);
      }

      if (sParam.eFileInfo) {
        sParam.content
          = lazy.UnMHTContentModifier.unmodifyContents(sParam.eFileInfo,
                                                       sParam.content,
                                                       sParam.charset);
      }

      UnMHTSaveUtils.modifyContent(sParam, sParam.savedFrom, this._saveDate,
                                   markerID);

      sParam.content
        = UnMHTDOMConverter.removeMarker(sParam.content, markerID);

      sParam.content
        = arUconv.fromUnicodeWithEntity(sParam.content, sParam.charset);

      sParam.status = UnMHTSaveStatus.DOC_CUR_GATHER_START;
      } catch (e) {
        Cu.reportError(e + "\n" + e.stack);
        if (UnMHTDocumentCategory.isDocumentAlive(sParam.targetDocument)) {
          Cu.reportError(e + "\n" + e.stack);
          sParam.reason = e.toString();
        } else {
          sParam.reason = TEXT("save.error.document_closed");
        }
        sParam.status = UnMHTSaveStatus.NG;
      }
    }, 0);
  },

  /**
   * ファイルを収集する
   *
   * @param   {UnMHTSaveParam} sParam
   *          対象の保存情報
   * @returns {Promise.<void>}
   * @resolves ダイアログ表示完了 (収集処理は終わっていない)
   */
  _startGatherDocument: function(sParam) {
    return new Promise((resolve, reject) => {
      sParam.treeWalker = new arTreeWalker(sParam.targetDocument.documentElement);

      Task.spawn(function*() {
        let saveVideo = false, saveEmbed = false;

        if (["video", "audio", "source"].some(name => sParam.targetDocument.getElementsByTagName(name).length > 0)) {
          saveVideo = yield* this._checkSaveContentsTask("video");
        }
        if (["embed", "object"].some(name => sParam.targetDocument.getElementsByTagName(name).length > 0)) {
          saveEmbed = yield* this._checkSaveContentsTask("embed");
        }

        resolve();

        let isRemoteGather = sParam.status == UnMHTSaveStatus.REMOTE_GATHER_WAIT;

        let targetWindow = sParam.targetDocument.defaultView;

        let checkRule = (rule, checkSheet) => {
          let IMPORT_RULE = 3;
          switch (rule.type) {
          case IMPORT_RULE: {
            checkSheet(rule.styleSheet);
            break;
          }
          }

          if (rule instanceof targetWindow.CSSGroupingRule) {
            for (let childRule of Array.from(rule.cssRules)) {
              checkRule(childRule, checkSheet);
            }
          }
        };

        let checkSheet = sheet => {
          if (!sheet) {
            /* Adblock 等によってスタイルシートがブロックされた場合 */
            return;
          }

          if (sheet.disabled) {
            return;
          }

          if (sheet.href) {
            this._checkAndPush(sheet.href, Ci.nsIContentPolicy.TYPE_STYLESHEET,
                               sParam.targetDocument, sParam.referrer,
                               null, isRemoteGather);
          }

          for (let rule of Array.from(sheet.cssRules)) {
            checkRule(rule, checkSheet);
          }
        };

        for (let sheet of Array.from(sParam.targetDocument.styleSheets)) {
          checkSheet(sheet);
        }

        let startTime = Date.now();
        let range = sParam.targetDocument.createRange();

        let gatherFont = node => {
          range.selectNode(node);
          let fontList = lazy.DOMUtils.getUsedFontFaces(range);
          for (let i = 0, length = fontList.length; i < length; i += 1) {
            let font = fontList.item(i);
            if (!font.URI) {
              continue;
            }
            this._checkAndPush(font.URI, Ci.nsIContentPolicy.TYPE_FONT,
                               sParam.targetDocument, sParam.referrer,
                               null,
                               isRemoteGather);
          }
        };

        gatherFont(sParam.targetDocument.documentElement);

        /* ループ内で使うプロパティのキャッシュ */
        let browser = sParam.browser;
        let targetDocument = sParam.targetDocument;
        let location = sParam.location;
        let savedFrom = sParam.savedFrom;
        let saveState = sParam.saveState;
        let baseDir = sParam.baseDir;
        let targetSelection = null;
        if (sParam.useSelection) {
          targetSelection = targetWindow.getSelection();
          if (!targetSelection ||
              targetSelection.isCollapsed ||
              targetSelection.rangeCount === 0) {
            targetSelection = null;
          }
        }
        let referrer = sParam.referrer;
        let isMail = sParam.isMail;
        let frameIDs = sParam.frameIDs;
        let treeWalker = sParam.treeWalker;

        let count = 0;
        for (let node of treeWalker.gen) {
          count += 1;
          if (count == 10) {
            count = 0;
            let now = Date.now();
            if (now > startTime + 100) {
              /* UI に処理を返すために 100ms 以上実行したら 10ms 待つ */
              yield arTaskUtils.wait(10);
              startTime = now;

              if (this._isCanceled) {
                throw UnMHTSaveResult.CANCEL;
              }
            }
          }

          if (node instanceof targetWindow.Element) {
            this._gatherFilesFromNode(
              browser,
              targetWindow,
              targetDocument,
              location, savedFrom, saveState, node,
              baseDir,
              treeWalker.isHead ? null: targetSelection,
              referrer, isMail,
              isRemoteGather,
              frameIDs, saveVideo, saveEmbed);
          }
        }

        switch (sParam.status) {
        case UnMHTSaveStatus.DOC_CUR_GATHER_WAIT: {
          sParam.status = UnMHTSaveStatus.OK;
          break;
        }
        case UnMHTSaveStatus.DOC_ORI_GATHER_WAIT: {
          sParam.status = UnMHTSaveStatus.FILE_NA;
          break;
        }
        case UnMHTSaveStatus.MAIL_ORI_GATHER_WAIT: {
          /* 完了したらこのパートは不要 */
          sParam.status = UnMHTSaveStatus.IGNORE;
          break;
        }
        case UnMHTSaveStatus.REMOTE_GATHER_WAIT: {
          /* 完了したらこのパートは不要 */
          sParam.status = UnMHTSaveStatus.IGNORE;
          break;
        }
        }
      }.bind(this)).catch(e => {
        if (e == UnMHTSaveResult.CANCEL) {
          sParam.reason = TEXT("save.error.canceled");
        } else if (UnMHTDocumentCategory.isDocumentAlive(sParam.targetDocument)) {
          Cu.reportError(e + "\n" + e.stack);
          sParam.reason = e;
        } else {
          sParam.reason = TEXT("save.error.document_closed");
        }
        sParam.status = UnMHTSaveStatus.NG;
      });
    });
  },

  /**
   * favicon の取得を開始する
   *
   * @param   {UnMHTSaveParam} sParam
   *          対象の保存情報
   */
  _startGetFavicon: function(sParam) {
    try {
      let pageURI = Services.io.newURI(sParam.savedFrom, null, null);

      let listener = {
        QueryInterface: XPCOMUtils.generateQI([Ci.nsIFaviconDataCallback]),

        /**
         * favicon のデータが到着したイベント
         *
         * @param   {nsIURI} uri
         *          favicon の URI
         * @param   {number} dataLen
         *          データの長さ
         * @param   {string} data
         *          データ
         * @param   {string} mimeType
         *          MIME-Type
         */
        onComplete: function(uri, dataLen, data, mimeType) {
          if (!uri) {
          } else if (uri.scheme == "data") {
            sParam.faviconData = uri.spec;
            if (mimeType) {
              sParam.faviconType = mimeType;
            } else {
              let m = sParam.faviconData.match(/^data:(image\/[^;,]+)/);
              if (m) {
                sParam.faviconType = m[1];
              }
            }
          } else if (dataLen && mimeType) {
            let bindata = fromUint8Array(new Uint8Array(data));
            sParam.faviconData
              = "data:" + mimeType + ";base64,"
              + arMIMEEncoder.encodeBase64(bindata, false);
            sParam.faviconType = mimeType;
          }

          if (sParam.status == UnMHTSaveStatus.DOC_CUR_FAVICON_WAIT) {
            sParam.status = UnMHTSaveStatus.DOC_CUR_SERIAL_START;
          } else if (sParam.status == UnMHTSaveStatus.DOC_ORI_FAVICON_WAIT) {
            sParam.status = UnMHTSaveStatus.DOC_ORI_GATHER_START;
          }
        }
      };

      lazy.FaviconService.getFaviconDataForPage(pageURI, listener);
    } catch (e) {
      if (sParam.status == UnMHTSaveStatus.DOC_CUR_FAVICON_WAIT) {
        sParam.status = UnMHTSaveStatus.DOC_CUR_SERIAL_START;
      } else if (sParam.status == UnMHTSaveStatus.DOC_ORI_FAVICON_WAIT) {
        sParam.status = UnMHTSaveStatus.DOC_ORI_GATHER_START;
      }
    }
  },

  /**
   * ファイルの取得を開始する
   *
   * @param   {UnMHTSaveParam} sParam
   *          対象の保存情報
   * @param   {number} now
   *          現在時刻
   */
  _startFileLoad: function(sParam, now) {
    sParam.startTime = now;
    sParam.lastRecvTime = now;
    sParam.status = UnMHTSaveStatus.FILE_GET;

    let uri = null;
    let referrer = null;

    try {
      uri = Services.io.newURI(sParam.currentLocation, null, null);
    } catch (e) {
      sParam.status = UnMHTSaveStatus.NG;
      sParam.reason = TEXT("save.error.invalid_url");
      return;
    }
    if (sParam.referrer) {
      try {
        referrer = Services.io.newURI(sParam.referrer, null, null);
      } catch (e) {
      }
    }
    sParam.loadContext
      = arNetFileLoader.createContext(uri, referrer, true, sParam,
                                      this._window);
    sParam.loadContext.start();
  },

  /**
   * メールの取得を開始する
   *
   * @param   {UnMHTSaveParam} sParam
   *          対象の保存情報
   */
  _startGetMail: Task.async(function*(sParam) {
    let result = yield arNetFileLoader.loadData(sParam.messageURI, null, true,
                                                this._window);

    let mailPart = null;
    if (result.status != arNetFileLoaderStatus.NG) {
      mailPart
        = arMIMEDecoder.decodeMessage(result.data.replace(/\r\n|\r|\n/g, "\r\n"),
                                      true);
    }

    if (mailPart &&
        UnMHTPref.getPref("save.header.mail.original_field")) {
      this._addAllFields(mailPart, true);
    }

    if (sParam.status == UnMHTSaveStatus.MAIL_CUR_MAIL_WAIT) {
      sParam.status = UnMHTSaveStatus.DOC_CUR_NA;
      this._mailPart = mailPart;
    } else if (sParam.status == UnMHTSaveStatus.MAIL_ORI_MAIL_WAIT) {
      if (mailPart) {
        for (let name of mailPart.fieldNames) {
          let nameL = name.toLowerCase();
          if (!(nameL == "content-type" ||
                nameL == "content-transfer-encoding")) {
            mailPart.removeField(name);
          }
        }

        this._topPart.parts.push(mailPart);

        sParam.status = UnMHTSaveStatus.MAIL_ORI_NA;
      } else {
        sParam.status = UnMHTSaveStatus.NG;
        sParam.reason = TEXT("save.error.email");
      }
    }
  })
});

/**
 * タブを MHT で保存
 *
 * @class
 */
let UnMHTSave = Object.freeze({
  /**
   * 保存コンテキストを作成する
   *
   * @param   {nsIDOMWindow} window
   *          実行中のウィンドウ
   * @param   {string} targetFilePath
   *          対象のファイルのパス
   * @param   {string} location
   *          メールを保存する際のルートドキュメントの URI
   * @param   {string} subject
   *          ファイルの subject
   * @param   {nsIDownloadProgressListener} downloadListener
   *          ダウンロードのイベントリスナ
   * @param   {number} saveState
   *          保存状態の数値表現
   * @returns {UnMHTSaveContext}
   *          保存コンテキスト
   */
  createContext: function(window, targetFilePath, location, subject,
                          downloadListener, saveState) {
    return new UnMHTSaveContext(window, targetFilePath, location, subject,
                                downloadListener, saveState);
  },

  /**
   * 同名のファイルが存在した場合の警告を表示する
   *
   * @param   {nsIDOMWindow} window
   *          実行中のウィンドウ
   * @param   {string} targetFilePath
   *          同名のファイル
   * @param   {string} base
   *          ファイル名の拡張子以外
   * @param   {boolean} showApplyToAll
   *          全てに適用チェックを表示するか
   * @param   {boolean} showSkip
   *          スキップを表示するか
   * @returns {Promise.<Object>}
   * @resolves 今回の動作
   */
  showOverwriteConfirm: Task.async(function*(window, targetFilePath, base,
                                             showApplyToAll, showSkip) {
    let parent = OS.Path.dirname(targetFilePath);
    let basename = OS.Path.basename(targetFilePath);

    let targetFileStat = yield OS.File.stat(targetFilePath);
    let lastModifiedDate = targetFileStat.lastModificationDate;

    let renamePath = yield this.getUniqueFilePath(parent, base);

    let arg = {
      title:    TEXT("save.overwrite.title"),
      message:  (TEXT("save.overwrite.message", [basename]) + "\n"
                 + TEXT("save.overwrite.lastaccess")
                 + " " + arFormatter.date(lastModifiedDate) + "\n"
                 + TEXT("save.overwrite.message2")),
      accept:   TEXT("replace"),
      extra1:   showSkip ? TEXT("skip") : null,
      extra2:   TEXT("abort"),
      rename:   TEXT("rename"),
      toall:    showApplyToAll ? TEXT("save.overwrite.all") : null,
      filePath: renamePath,
      exist:    name => TEXT("save.overwrite.message", [name]),
      focus:    "yes"
    };

    let result = yield lazy.UnMHTConfirmDialog.open(window, arg);

    return result;
  }),

  /**
   * ユニークなファイルを返す
   *
   * @param   {string} parent
   *          親ディレクトリのパス
   * @param   {string} base
   *          ファイル名の拡張子以外
   * @returns {Promise.<string>}
   * @resolves ユニークなファイルのパス
   */
  getUniqueFilePath: Task.async(function*(parent, base) {
    let options = {
      allowExists: false,
      noCreate: true,
      indexInit: UnMHTPref.getPref("save.filename.dup.rename.init"),
      indexDigits: UnMHTPref.getPref("save.filename.dup.rename.digits")
    };
    let suffix = "." + UnMHTFilename.defaultExtension;
    let result = yield arFileUtils.prepareFile(parent, base, suffix,
                                               options);

    return result.path;
  }),

  /**
   * 保存する際のファイル名を取得する
   *
   * @param   {nsIDOMDocument} targetDocument
   *          対象のドキュメント
   * @param   {Object} scrapBookData
   *          ScrapBook のデータ
   * @param   {string} filenameType
   *          ファイル名の種類
   *            title/leafname/custom
   * @param   {Date} date
   *          日時
   *          カスタムファイル名に使用する
   * @returns {string}
   *          ファイル名 (拡張子以外)
   */
  getFilenameBase: function(targetDocument, scrapBookData, filenameType,
                            date) {
    let [uri, title]
      = this.getLocationAndTitle(targetDocument, scrapBookData);

    let decodedURI = uri;
    try {
      decodedURI = decodeURI(uri);
    } catch (e) {
    }

    let [path, fragment] = arPathUtils.splitFragment(uri);
    let m = path.match(/^([^\?]*)\?([^\?]*)$/);
    if (m) {
      path = m[1];
    }
    let decodedPath = decodedURI;
    try {
      decodedPath = decodeURI(path);
    } catch (e) {
    }

    let localFilePath = { value: null };
    if (UnMHTPref.getPref("save.local.same_dir") &&
        UnMHTDocumentCategory.isLocalFile(decodedURI, localFilePath)) {
      return OS.Path.basename(localFilePath.value).replace(/\.[^\.]+$/, "")
        + UnMHTPref.getPref("save.local.same_dir.append");
    }

    let filename = "";

    m = decodedPath.match(/\/([^\/]+)$/);
    if (m) {
      filename = m[1].replace(/\.[^\.]+$/, "");
    } else {
      filename = "index";
    }
    title = title || filename;

    if (!filenameType) {
      filenameType = UnMHTPref.getPref("save.filename");
    }
    if (UnMHTMailURLUtils.isMailURL(decodedURI) &&
        filenameType == "filename") {
      filenameType = "title";
    }

    switch (filenameType) {
      case "title": {
        filename = UnMHTFilename.validateFileName(title) || filename;
        break;
      }
      case "custom": {
        filename
          = UnMHTFilename.validateFileName(lazy.UnMHTCustomFilename.format(title, filename, uri, date));
        break;
      }
      default: {
        filename = UnMHTFilename.validateFileName(filename);
      }
    }

    return filename;
  },

  /**
   * ファイル名中の使用不可能な文字の置換、
   * 長すぎるファイル名の短縮、拡張子の追加をする
   *
   * @param   {string} origFilePath
   *          対象のファイルのパス
   * @returns {Promise.<string>}
   * @resolves 保存可能なファイルのパス
   * @rejects UnMHTSaveResult.CANCEL
   */
  getValidFilePath: Task.async(function*(window, origFilePath) {
    let origBasename = OS.Path.basename(origFilePath);
    let parent = OS.Path.dirname(origFilePath);
    let validBasename = UnMHTFilename.validateFileName(origBasename);
    let validPath = OS.Path.join(parent, validBasename);

    if (UnMHTFilename.removeExtension(validBasename) == validBasename) {
      /* 拡張子 が無いと UnMHT でハンドルできなくなるので追加する */
      let suffix = "." + UnMHTFilename.defaultExtension;
      let options = { allowExists: true, noCreate: true };
      let result = yield arFileUtils.prepareFile(parent, validBasename,
                                                 suffix, options);
      if (result.exists) {
        /* .mht を追加した事でファイル名が被った */
        let oResult = yield this.showOverwriteConfirm(window, result.path,
                                                      validBasename,
                                                      false, false);

        if (oResult.action == "abort" || oResult.action == "no") {
          throw UnMHTSaveResult.CANCEL;
        } else if (oResult.action == "rename") {
          validPath = oResult.renamePath;
        } else {
          validPath = result.path;
        }
      }
    }

    return validPath;
  }),

  /**
   * ドキュメントのアドレスとタイトルを取得する
   *
   * @param   {?nsIDOMDocument} targetDocument
   *          対象のドキュメント
   * @param   {?Object} scrapBookData
   *          ScrapBook のデータ
   * @returns Array
   *            [stirng アドレス, string タイトル]
   */
  getLocationAndTitle: function(targetDocument, scrapBookData) {
    if (scrapBookData) {
      return [scrapBookData["source"], scrapBookData["title"]];
    } else {
      return [targetDocument.documentURI, targetDocument.title];
    }
  },

  /**
   * 追加 URI を解析する
   *
   * @param   {string} text
   *          行ごとに URI を記述したテキスト
   * @returns {Array.<string>}
   *          URI リスト
   */
  parseAdditionalURIs: function(text) {
    return text.split(/[\r\n]/)
      .map(line => line.trim())
      .filter(uri => uri);
  }
});

module.UnMHTSave = UnMHTSave;
module.UnMHTSaveUtils = UnMHTSaveUtils;
module.UnMHTSaveResult = UnMHTSaveResult;

})(this);
