"use strict";

this.EXPORTED_SYMBOLS = ["arAlertUtils"];

(function(module) {

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
let { Promise } = Cu.import("resource://gre/modules/Promise.jsm", {});
let { Services } = Cu.import("resource://gre/modules/Services.jsm", {});
let { setTimeout } = Cu.import("resource://gre/modules/Timer.jsm", {});

/**
 * 警告関連ユーティリティ
 *
 * @class
 */
let arAlertUtils = Object.freeze({
  /**
   * Android の toast alert を表示する
   *
   * @param   {nsIDOMWindow} window
   *          実行中のウィンドウ
   * @param   {string} message
   *          表示するメッセージ
   * @param   {string} duration
   *          表示時間
   */
  toast: function(window, message, duration) {
    window.NativeWindow.toast.show(message, duration);
  },

  /**
   * Android の doorhanger notification を表示する
   *
   * @param   {nsIDOMWindow} window
   *          実行中のウィンドウ
   * @param   {string} message
   *          表示するメッセージ
   * @param   {string} value
   *          タグ
   * @param   {Array<Object>} buttons
   *          ボタン情報
   * @param   {Object} options
   *          オプション
   */
  doorhanger: function(window, message, value, buttons, options) {
    window.NativeWindow.doorhanger.show(message, value, buttons,
                                        window.BrowserApp.selectedTab.id,
                                        options);
  },

  /**
   * 非同期で確認を表示する
   *
   * @param   {nsIDOMWindow} window
   *          実行中のウィンドウ
   * @param   {string} title
   *          タイトル
   * @param   {string} message
   *          表示するメッセージ
   * @returns {Promise.<boolean>}
   * @resolves 結果
   */
  confirm: function(window, title, message) {
    return new Promise(function(resolve, reject) {
      setTimeout(() => {
        let ret = Services.prompt.confirm(window, title, message);

        resolve(ret);
      }, 0);
    });
  },

  /**
   * 非同期でアラートを表示する
   *
   * @param   {nsIDOMWindow} window
   *          実行中のウィンドウ
   * @param   {string} title
   *          タイトル
   * @param   {string} message
   *          表示するメッセージ
   * @returns {Promise.<void>}
   * @resolves 終了
   */
  alert: function(window, title, message) {
    return new Promise(function(resolve, reject) {
      setTimeout(() => {
        Services.prompt.alert(window, title, message);

        resolve();
      }, 0);
    });
  }
});

module.arAlertUtils = arAlertUtils;

})(this);
