"use strict";

this.EXPORTED_SYMBOLS = ["arArrayUtils"];

(function(module) {

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;

/**
 * 配列ユーティリティ
 *
 * @class
 */
let arArrayUtils = Object.freeze({
  /**
   * 配列 a の要素 x に f を適用した y について
   * 最初に pred を満たした y を返す
   * どの y も pred を満たさない場合は undefined を返す
   *
   * 速度が必要な場面で map と filter の代わりに使う
   * f と p が副作用を持たない場合に限り下の 2 つは等価
   * [y] = a.map(f).filter(p);
   * y = arArrayUtils.mapfind(a, f, p);
   *
   * @template ElemType
   * @template RetType
   * @param   {Array.<ElemType>} a
   *          対象の配列
   * @param   {function(ElemType,number,Array.<ElemType>):RetType} f
   *          a の要素に適用する関数
   *            function(item, index, array)
   *              *param   {ElemType} item
   *              *param   {number} index
   *              *param   {Array.<ElemType>} array
   *              *returns {RetType}
   * @param   {function(RetType,number,Array.<ElemType>):boolean} pred
   *          述語
   *            function(item, index, array)
   *              *param   {RetType} item
   *              *param   {number} index
   *              *param   {Array.<ElemType>} array
   *              *returns {boolean}
   * @returns {(RetType|undefined)}
   *          f(a の要素) もしくは undefined
   */
  mapfind: function(a, f, pred) {
    for (let i = 0, length = a.length; i < length; i += 1) {
      let y = f(a[i], i, a);
      if (pred(y, i, a)) {
        return y;
      }
    }

    return undefined;
  },

  /**
   * [begin, end) のジェネレータ
   *
   * @param   {number} begin
   *          始値
   * @param   {number} end
   *          終値
   * @returns {Generator}
   *          [begin, end) のジェネレータ
   */
  range: function*(begin, end) {
    if (begin < end) {
      for (let i = begin; i < end; i += 1) {
        yield i;
      }
    } else {
      for (let i = begin; i > end; i -= 1) {
        yield i;
      }
    }
  }
});

module.arArrayUtils = arArrayUtils;

})(this);
