"use strict";

this.EXPORTED_SYMBOLS = ["arCustomizableUI"];

(function(module) {

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
let { arShutdownEvent } = Cu.import("resource://unmht/modules/base/arShutdownEvent.jsm", {});

let CustomizableUI = null;
try {
  CustomizableUI = Cu.import("resource:///modules/CustomizableUI.jsm", {}).CustomizableUI;
} catch (e) {
}

/**
 * CustomizableUI のラッパー
 *
 * @class
 */
let arCustomizableUI = Object.freeze({
  /**
   * 定数取得用の CustomizableUI
   * @type {CustomizableUI}
   */
  CustomizableUI: CustomizableUI,

  /**
   * CustomizableUI がサポートされているか
   * @type {boolean}
   */
  supported: CustomizableUI !== null,

  /**
   * ウィジェットを作成する
   * シャットダウン時には自動で破棄する
   * 追加済みのウィジェットは追加しない
   *
   * @param   {Object} params
   *          CustomizableUI.createWidget の引数
   */
  createWidget: function(params) {
    if (this._ids.has(params.id)) {
      return;
    }
    CustomizableUI.createWidget(params);
    this._ids.add(params.id);
  },

  /**
   * バッチ更新
   */
  batch: function(f) {
    CustomizableUI.beginBatchUpdate();
    f();
    CustomizableUI.endBatchUpdate();
  },

  /* ==== arIShutdownEventListener ==== */

  /**
   * シャットダウンイベント
   *
   * @param   {Object} data
   *          Bootstrap データ
   * @param   {number} reason
   *          理由
   */
  onShutdown: function(data, reason) {
    this.batch(() => {
      for (let id of this._ids) {
        CustomizableUI.destroyWidget(id);
      }
    });
    this._ids.clear();
  },

  /* ==== private ==== */

  /**
   * 登録したウィジェットの ID
   * @type {Set.<string>}
   */
  _ids: new Set(),

  /**
   * 初期化処理
   */
  init: function() {
    if (this.supported) {
      arShutdownEvent.addListener(this);
    }
  }
});
arCustomizableUI.init();

module.arCustomizableUI = arCustomizableUI;

})(this);
