"use strict";

this.EXPORTED_SYMBOLS = ["arFennecMenu"];

(function(module) {

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
let { arScoped } = Cu.import("resource://unmht/modules/base/arScoped.jsm", {});

/**
 * ウィンドウごとのメニュー管理
 *
 * @param   {nsIDOMWindow} window
 *          対象のウィンドウ
 * @constructor
 * @implements arIScopedClass
 */
function arFennecMenuPerWindow(window) {
  /**
   * 対象のウィンドウ
   * @type {nsIDOMWindow}
   */
  this._window = window;

  /**
   * 追加したメニューの ID
   * @type {Set.<arFennecMenu>}
   */
  this._ids = new Set();

  Object.seal(this);
}
arFennecMenuPerWindow.prototype = Object.freeze({
  /**
   * メニューを追加する
   *
   * @param   {Object} attrs
   *          属性
   * @param   {function} oncommand
   *          メニュークリック時のイベントハンドラ
   */
  add: function(attrs, oncommand) {
    let NativeWindow = this._window.NativeWindow;
    let id = NativeWindow.menu.add(attrs.label,
                                   ("icon" in attrs) ? attrs.icon : "",
                                   oncommand);

    this._ids.add(id);
  },

  /* ==== arIScopedClass ==== */

  /**
   * 破棄する
   *
   * @param   {boolean} isShutdown
   *          シャットダウンイベントか
   * @param   {Object} data
   *          Bootstrap データ
   * @param   {number} reason
   *          理由
   */
  destruct: function(isShutdown, data, reason) {
    for (let id of this._ids) {
      this._window.NativeWindow.menu.remove(id);
    }
    this._ids.clear();
    this._window = null;
  }
});

/**
 * メニュー管理
 *
 * @class
 */
let arFennecMenu = Object.freeze({
  /**
   * メニューを追加する
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   * @param   {Object} attrs
   *          属性
   * @param   {function} oncommand
   *          メニュークリック時のイベントハンドラ
   */
  add: function(window, attrs, oncommand) {
    this._menus.create(window).add(attrs, oncommand);
  },

  /* ==== private ==== */

  /**
   * ウィンドウごとのメニュー管理
   * @type {arScoped.<arFennecMenuPerWindow>}
   */
  _menus: new arScoped(arFennecMenuPerWindow)
});

module.arFennecMenu = arFennecMenu;

})(this);
