"use strict";

this.EXPORTED_SYMBOLS = ["arFormatter"];

(function(module) {

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
let { XPCOMUtils } = Cu.import("resource://gre/modules/XPCOMUtils.jsm", {});

let lazy = {};
XPCOMUtils.defineLazyServiceGetter(lazy, "ScriptableDateFormat",
                                   "@mozilla.org/intl/scriptabledateformat;1",
                                   "nsIScriptableDateFormat");

/**
 * テキストフォーマッタ
 *
 * @class
 */
let arFormatter = Object.freeze({
  /**
   * ファイルサイズを "," で区切る
   *
   * @param   {number} size
   *          ファイルサイズ
   * @returns {string}
   *          "," で区切られたファイルサイズ
   */
  size: function(size) {
    let readableSize = size.toString();

    if (readableSize.length > 3) {
      readableSize
        = this.size(readableSize.slice(0, -3))
        + "," + readableSize.slice(-3);
    }

    return readableSize;
  },

  /**
   * 日時を可読形式に変換する
   *
   * @param   {Date} date
   *          日時
   * @returns {string}
   *          日時の可読形式
   */
  date: function(date) {
    return lazy.ScriptableDateFormat
      .FormatDateTime("",
                      Ci.nsIScriptableDateFormat.dateFormatLong,
                      Ci.nsIScriptableDateFormat.timeFormatSeconds,
                      date.getFullYear(),
                      date.getMonth() + 1,
                      date.getDate(),
                      date.getHours(),
                      date.getMinutes(),
                      date.getSeconds());
  },

  /**
   * 指定桁になるまで 0 で埋める
   *
   * @param   {number} width
   *          桁数
   * @param   {number} n
   *          対象の数字
   * @returns {string}
   *          埋めた数字
   */
  pad: function(width, n) {
    let s = n + "";
    return "0".repeat(Math.max(width - s.length, 0)) + s;
  }
});

module.arFormatter = arFormatter;

})(this);
