"use strict";

this.EXPORTED_SYMBOLS = ["arPrototype"];

(function(module) {

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;

/**
 * プロトタイプの継承を簡略化する
 *
 * @class
 */
let arPrototype = Object.freeze({
  /**
   * 継承したプロトタイプを作成する
   *
   * @param   {Constructor} child
   *          子
   * @param   {Constructor} parent
   *          親
   */
  extend: function(child, parent) {
    let prop = {};
    for (let name of Object.getOwnPropertyNames(child.prototype)) {
      prop[name] = Object.getOwnPropertyDescriptor(child.prototype, name);
    }
    child.prototype = Object.create(parent.prototype, prop);
    if (Object.isFrozen(parent.prototype)) {
      Object.freeze(child.prototype);
    } else if (Object.isSealed(parent.prototype)) {
      Object.seal(child.prototype);
    }
  }
});

module.arPrototype = arPrototype;

})(this);
