"use strict";

this.EXPORTED_SYMBOLS = ["arShutdownEvent"];

(function(module) {

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;

/**
 * シャットダウンイベント通知
 *
 * @class
 */
let arShutdownEvent = Object.freeze({
  /**
   * イベントリスナを追加する
   * オブジェクトのクリーンアップや、
   * 追加と削除が対にならないものに使用する
   *
   * @param   {arIShutdownEventListener} listener
   *          イベントリスナ
   */
  addListener: function(listener) {
    this._listeners.add(listener);
  },

  /**
   * 関数のイベントリスナを追加する
   * 追加と削除などの対になっているものの削除側や一回限りのものに使用する
   *
   * @param   {function(Object,number)} func
   *          関数のイベントリスナ
   *            *param   {Object} data
   *                     Bootstrap データ
   *            *param   {number} reason
   *                     理由
   */
  addFunc: function(func) {
    this._funcs.add(func);
  },

  /**
   * シャットダウンイベントを通知する
   *
   * @param   {Object} data
   *          Bootstrap データ
   * @param   {number} reason
   *          理由
   */
  broadcast: function(data, reason) {
    for (let listener of this._listeners) {
      try {
        listener.onShutdown(data, reason);
      } catch (e) {
        Cu.reportError(e + "\n"
                       + e.fileName + ":" + e.lineNumber + "\n"
                       + e.stack);
      }
    }
    this._listeners.clear();

    for (let func of this._funcs) {
      try {
        func(data, reason);
      } catch (e) {
        Cu.reportError(e + "\n"
                       + e.fileName + ":" + e.lineNumber + "\n"
                       + e.stack);
      }
    }
    this._funcs.clear();
  },

  /* ==== private ==== */

  /**
   * イベントリスナ
   * @type {Set.<arIShutdownEventListener>}
   */
  _listeners: new Set(),

  /**
   * 関数のイベントリスナ
   * @type {Set.<function(Object,number)>}
   */
  _funcs: new Set()
});

module.arShutdownEvent = arShutdownEvent;

})(this);
