"use strict";

this.EXPORTED_SYMBOLS = ["arStringBundle"];

(function(module) {

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
let { Services } = Cu.import("resource://gre/modules/Services.jsm", {});
let { arPref } = Cu.import("resource://unmht/modules/base/arPrefClass.jsm", {});
let { arShutdownEvent } = Cu.import("resource://unmht/modules/base/arShutdownEvent.jsm", {});

const LOCALE_PREF = "general.useragent.locale";

/**
 * StringBundle 管理
 *
 * @constructor
 * @param   {string} uri
 *          バンドルの URI
 * @implements arIShutdownEventListener
 */
function arStringBundle(uri) {
  /**
   * バンドルの URI
   * @type {string}
   */
  this._uri = uri;

  /**
   * StringBundle
   * @type {nsIStringBundle}
   */
  this._bundle = null;

  arShutdownEvent.addListener(this);

  arPref.addGlobalListener(LOCALE_PREF, () => this._updateBundle(), true);

  Object.seal(this);
}
arStringBundle.prototype = Object.freeze({
  /**
   * StringBundle から文字列を取得する
   *
   * @param   {string} name
   *          取得する文字列の名前
   * @param   {Array=} params
   *          フォーマット用のパラメータ (オプショナル)
   * @returns {string}
   *          StringBundle の文字列
   */
  get: function(name, params) {
    try {
      if (params) {
        return this._bundle.formatStringFromName(name, params, params.length);
      } else {
        return this._bundle.GetStringFromName(name);
      }
    } catch (e) {
      Cu.reportError(name + "," + params + "\n" + e + "\n" + e.stack);
      return "";
    }
  },

  /**
   * キャッシュをクリアする
   */
  flushBundle: function() {
    Services.strings.flushBundles();
  },

  /* ==== arIShutdownEventListener ==== */

  /**
   * シャットダウンイベント
   *
   * @param   {Object} data
   *          Bootstrap データ
   * @param   {number} reason
   *          理由
   */
  onShutdown: function(data, reason) {
    this.flushBundle();
  },

  /* ==== private ==== */

  /**
   * バンドルを更新する
   */
  _updateBundle: function() {
    this._bundle = Services.strings.createBundle(this._uri);
  }
});

module.arStringBundle = arStringBundle;

})(this);
