"use strict";

this.EXPORTED_SYMBOLS = ["arTaskUtils"];

(function(module) {

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
let { Promise } = Cu.import("resource://gre/modules/Promise.jsm", {});
let { setTimeout, clearTimeout } = Cu.import("resource://gre/modules/Timer.jsm", {});

/**
 * Task.jsm 用ユーティリティ
 *
 * @class
 */
let arTaskUtils = Object.freeze({
  TIMEOUT: 0x0001,
  CLEARED: 0x0002,

  /**
   * 指定時間後に解決する Promise を返す
   *
   * @param   {number} t
   *          待つ時間 [ms]
   * @param   {object} timerState
   *          (出力) タイマーの状態
   * @returns {Promise.<number>}
   * @resolves arTaskUtils.TIMEOUT: 指定時間待った
   *           arTaskUtils.CLEARED: 解除された
   */
  wait: function(t, timerState) {
    return new Promise((resolve, reject) => {
      let _timer = setTimeout(() => {
        resolve(this.TIMEOUT);
      }, t);
      if (timerState) {
        timerState.timer = _timer;
        timerState.reject = reject;
      }
    });
  },

  /**
   * タイマーを解除する
   *
   * @param   {object} timerState
   *          (出力) タイマーの状態
   */
  clear: function(timerState) {
    if (timerState &&
        "timer" in timerState && timerState.timer &&
        "reject" in timerState && timerState.reject) {
      clearTimeout(timerState.timer);
      timerState.reject(this.CLEARED);
      timerState.timer = null;
      timerState.reject = null;
    }
  }
});

module.arTaskUtils = arTaskUtils;

})(this);
