"use strict";

this.EXPORTED_SYMBOLS = ["arTreeWalker"];

(function(module) {

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;

/**
 * DOM ツリー走査
 *
 * @constructor
 * @param   {nsIDOMElement} startNode
 *          開始要素
 */
function arTreeWalker(startNode) {
  /**
   * HEAD 要素内か
   * @type {boolean}
   */
  this.isHead = false;

  /**
   * DOM ツリー走査ジェネレータ
   * @type {Generator}
   */
  this.gen = this._getGenerator(startNode);

  /**
   * 対象にする要素の数
   * @type {number}
   */
  this._nodeCount = 1;

  /**
   * 走査した要素の数
   * @type {number}
   */
  this._walked = 0;

  Object.seal(this);
}
arTreeWalker.prototype = Object.freeze({
  /**
   * 進捗状況 (0-1)
   * 走査した HTMLElement の数の割合
   *   @type {number}
   */
  get progress() {
    let p = this._walked / this._nodeCount;
    if (p > 1) {
      return 1;
    }
    return p;
  },

  /**
   * 破棄
   */
  destruct: function() {
    if (this.gen) {
      this.gen.close();
      this.gen = null;
    }
  },

  /* ==== private ==== */

  /**
   * 要素を返すジェネレータを作成する
   *
   * @param   {nsIDOMElement} startNode
   *          開始要素
   * @returns {Generator}
   *          要素を返すジェネレータ
   */
  _getGenerator: function*(startNode) {
    let targetWindow = startNode.ownerDocument.defaultView;

    let node = startNode;
    this.isHead = false;

    let i = 0;
    try {
      this._nodeCount = startNode.getElementsByTagName("*").length + 1;
    } catch (e) {
    }
    while (node) {
      if (node instanceof targetWindow.HTMLHeadElement) {
        this.isHead = true;
      }

      this._walked = i;

      yield node;
      if (node instanceof targetWindow.Element) {
        i += 1;
      }

      if (node.firstChild) {
        node = node.firstChild;
      } else {
        while (node) {
          if (node.nextSibling) {
            if (node instanceof targetWindow.HTMLHeadElement) {
              this.isHead = false;
            }

            node = node.nextSibling;
            break;
          }
          node = node.parentNode;
          if (node == startNode) {
            this._walked = this._nodeCount;
            return;
          }
        }
      }
    }
  }
});

module.arTreeWalker = arTreeWalker;

})(this);
