"use strict";

this.EXPORTED_SYMBOLS = ["startup"];

(function(module) {

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
let { Services } = Cu.import("resource://gre/modules/Services.jsm", {});
let { Task } = Cu.import("resource://gre/modules/Task.jsm", {});
let { XPCOMUtils } = Cu.import("resource://gre/modules/XPCOMUtils.jsm", {});
let { arAppInfo } = Cu.import("resource://unmht/modules/base/arAppInfo.jsm", {});
let { arWindowUtils } = Cu.import("resource://unmht/modules/base/arWindowUtils.jsm", {});
let { UnMHTPref } = Cu.import("resource://unmht/modules/UnMHTPref.jsm", {});

let lazy = {};
XPCOMUtils.defineLazyModuleGetter(lazy, "arComponentLoader",
                                  "resource://unmht/modules/base/arComponentLoader.jsm");
XPCOMUtils.defineLazyModuleGetter(lazy, "arStylesheetLoader",
                                  "resource://unmht/modules/base/arStylesheetLoader.jsm");

XPCOMUtils.defineLazyModuleGetter(lazy, "UnMHTContentPolicy",
                                  "resource://unmht/modules/UnMHTContentPolicy.jsm");
XPCOMUtils.defineLazyModuleGetter(lazy, "UnMHTContentSniffer",
                                  "resource://unmht/modules/UnMHTContentSniffer.jsm");
XPCOMUtils.defineLazyModuleGetter(lazy, "UnMHTDocumentLoaderFactory",
                                  "resource://unmht/modules/UnMHTDocumentLoaderFactory.jsm");
XPCOMUtils.defineLazyModuleGetter(lazy, "UnMHTProtocolHandler",
                                  "resource://unmht/modules/UnMHTProtocolHandler.jsm");

XPCOMUtils.defineLazyModuleGetter(lazy, "UnMHTFennecMenu",
                                  "resource://unmht/modules/UnMHTFennecMenu.jsm");

XPCOMUtils.defineLazyModuleGetter(lazy, "UnMHTMenu",
                                  "resource://unmht/modules/UnMHTMenu.jsm");
XPCOMUtils.defineLazyModuleGetter(lazy, "UnMHTLocationBar",
                                  "resource://unmht/modules/UnMHTLocationBar.jsm");
XPCOMUtils.defineLazyModuleGetter(lazy, "UnMHTToolbar",
                                  "resource://unmht/modules/UnMHTToolbar.jsm");
XPCOMUtils.defineLazyModuleGetter(lazy, "UnMHTKeyboardShortcut",
                                  "resource://unmht/modules/UnMHTKeyboardShortcut.jsm");

XPCOMUtils.defineLazyModuleGetter(lazy, "UnMHTPageInfo",
                                  "resource://unmht/modules/UnMHTPageInfo.jsm");

XPCOMUtils.defineLazyModuleGetter(lazy, "UnMHTFileLoadHook",
                                  "resource://unmht/modules/UnMHTFileLoadHook.jsm");
XPCOMUtils.defineLazyModuleGetter(lazy, "UnMHTDLFRegisterer",
                                  "resource://unmht/modules/UnMHTDLFRegisterer.jsm");

XPCOMUtils.defineLazyModuleGetter(lazy, "UnMHTMAFIntegration",
                                  "resource://unmht/modules/UnMHTMAFIntegration.jsm");
XPCOMUtils.defineLazyModuleGetter(lazy, "UnMHTScrapBookIntegration",
                                  "resource://unmht/modules/UnMHTScrapBookIntegration.jsm");

XPCOMUtils.defineLazyModuleGetter(lazy, "UnMHTFennecPrefPage",
                                  "resource://unmht/modules/UnMHTFennecPrefPage.jsm");

/**
 * 有効な場合の起動時、無効から有効への変更時 (更新含む)
 *
 * @param   {Object} data
 *          Bootstrap データ
 * @param   {number} reason
 *          理由
 *            1: APP_STARTUP
 *            3: ADDON_ENABLE
 *            5: ADDON_INSTALL
 *            7: ADDON_UPGRADE
 *            8: ADDON_DOWNGRADE
 * @param   {string} version
 *          バージョン
 */
function startup(data, reason, version) {
  try {
    if (Services.appinfo.browserTabsRemoteAutostart) {
      // Disable on e10s.
      Cu.reportError("UnMHT is not compatible with multi-process. Disabled automatically.");
      return;
    }
  } catch (e) {
  }

  UnMHTPref.version = version;

  let startupPromises = [];

  if (!arAppInfo.isThunderbird) {
    let p;

    p = lazy.arComponentLoader.loadComponent(lazy.UnMHTDocumentLoaderFactory);
    startupPromises.push(p);
    p = lazy.arComponentLoader.loadComponent(lazy.UnMHTContentPolicy);
    startupPromises.push(p);
    p = lazy.arComponentLoader.loadComponent(lazy.UnMHTProtocolHandler);
    startupPromises.push(p);
  }
  if (arAppInfo.isSeaMonkey) {
    let p = lazy.arComponentLoader.loadComponent(lazy.UnMHTContentSniffer);
    startupPromises.push(p);
  }

  if (arAppInfo.isFennec) {
    lazy.UnMHTFennecMenu.onStartup(data, reason);
    lazy.UnMHTFennecPrefPage.onStartup(data, reason);
  } else {
    lazy.arStylesheetLoader.load("chrome://unmht/content/ui.css",
                                 Ci.nsIStyleSheetService.AGENT_SHEET);

    lazy.UnMHTMenu.onStartup(data, reason);
    lazy.UnMHTLocationBar.onStartup(data, reason);
    lazy.UnMHTToolbar.onStartup(data, reason);
    lazy.UnMHTKeyboardShortcut.onStartup(data, reason);

    lazy.UnMHTPageInfo.onStartup(data, reason);
  }

  if (!arAppInfo.isThunderbird) {
    Task.spawn(function*() {
      yield lazy.UnMHTDLFRegisterer.register(data, reason);
      for (let p of startupPromises) {
        yield p;
      }
      lazy.UnMHTFileLoadHook.onStartup(data, reason);
    }).catch(function(e) {
      Cu.reportError(e + "\n" + e.stack);
    });
  }

  if (!arAppInfo.isThunderbird && !arAppInfo.isFennec) {
    if (UnMHTPref.getPref("integrataion.maf")) {
      lazy.UnMHTMAFIntegration.onStartup(data, reason);
    }
    if (UnMHTPref.getPref("integrataion.scrapbook")) {
      lazy.UnMHTScrapBookIntegration.onStartup(data, reason);
    }
  }

  arWindowUtils.closeOnShutdown("chrome://unmht/");
}

module.startup = startup;

})(this);
